# coding: utf-8

import argparse
import getpass
import os
import sys
from sys import platform

import logging

# pylint: disable=wrong-import-position
from tionix_vdi_client import MODULE_VERSION, TRACEBACK_ENABLED  # nopep8

if not TRACEBACK_ENABLED:
    sys.tracebacklimit = 0
# import settings before run_diagnostic to allow logging in settings
from tionix_vdi_client.settings import (  # nopep8
    SessionSettings, CONF)

from tionix_vdi_client import exceptions  # nopep8
from tionix_vdi_client.i18n import ugettext as _  # nopep8
from tionix_vdi_client.run_self_diagnostics import run_diagnostics  # nopep8
from tionix_vdi_client.smart_card.exceptions import (
    SmartCardNotConfigured)  # nopep8
from tionix_vdi_client.ui_tk import TionixVDIClientController  # nopep8
from tionix_vdi_client.vdi_connect import (
    vdi_connect, VdiClient, rdp_connect)  # nopep8


logger = logging.getLogger('tionix_vdi_client')


class SmartCardSession(object):

    def __init__(self):
        self.login = self.password = self.session = None

    def __enter__(self):
        try:
            from tionix_vdi_client.smart_card import api  # nopep8
        except ImportError as err:
            logger.warning(u"Environment is not configured: %s. "
                           u"Couldn't retrieve login/pass from smart card",
                           err)
        else:
            self.login, self.password, self.session = api.get_session_data()

        return self.login, self.password, self.session

    def __exit__(self, exc_type, exc_value, traceback):
        if self.session:
            self.session.closeSession()


def check_required_parameters():
    return all((CONF.domain_name, CONF.cloud or CONF.secondary_cloud))


def main():
    exit_code = 0
    if platform == "win32":
        TionixVDIClientController().show_main_window()
    else:
        args = create_parser().parse_args()
        if args.diagnostic:
            debug = args.debug
            if debug:
                del sys.tracebacklimit
            run_diagnostics(debug=debug)
        elif args.no_ui:
            exit_code = connect_without_ui(args)
        else:
            try:
                exit_code = connect_with_smartcard(args)
            except SmartCardNotConfigured:
                controller = TionixVDIClientController()
                controller.show_main_window(ip_address=args.ikecfg)
    return exit_code


def connect_with_smartcard(args):
    project = args.project
    domain = args.domain or CONF.domain_name
    server_uri = args.server_uri or CONF.cloud
    connect = args.connect or CONF.connect

    with SmartCardSession() as wrapper:
        username, password, session = wrapper

        controller = TionixVDIClientController(card_session=session)

        if not check_required_parameters():
            controller.show_main_window(
                ip_address=args.ikecfg, login=username,
                password=password)

        conn_params = {
            'user': username,
            'password': password or '',
            'domain_name': domain,
        }

        if not password:
            reason = _(
                u'No saved password found, please enter a new password.')
            password = controller.show_change_password_window(
                reason=reason, conn_params=conn_params,
                context={'username': conn_params['user']}
            )
            if not password:
                return getattr(os, 'EX_SOFTWARE', 1)
            conn_params['password'] = password

        if CONF.project and project:
            conn_params['project'] = project

        CONF.cloud = server_uri
        CONF.connect = connect

        if args.ikecfg:
            conn_params['ip_address'] = args.ikecfg

        try:
            vdi_client = VdiClient(**conn_params)
            controller.vdi_client = vdi_client
            vm_ip = controller.show_connect_window(
                context={'username': conn_params['user']}
            )
            if vm_ip:
                rdp_params = dict(
                    user=vdi_client.auth_params['user'],
                    password=vdi_client.auth_params['password'],
                    domain_name=vdi_client.params.get('user_domain_name', ''),
                    vm_ip=vm_ip)

                rdp_connect(**rdp_params)
        except Exception as err:
            exceptions.process_exception(err, fall=True)

    return getattr(os, 'EX_OK', 0)


def connect_without_ui(args):
    session_settings = None

    username = args.username
    password = args.password

    if not (username and password):
        session_settings = SessionSettings()
        session_settings.read()
        username = username or session_settings.user
        password = password or session_settings.password

    project = args.project
    domain = args.domain or CONF.domain_name
    server_uri = args.server_uri or CONF.cloud
    connect = args.connect or CONF.connect

    conn_credentials = {
        'user': username,
        'password': password,
        'domain_name': domain,
    }

    if CONF.project:
        conn_credentials['project'] = project

    if args.store_session:
        if not session_settings:
            session_settings = SessionSettings()
            session_settings.read()

        session_settings.update(conn_credentials)
        session_settings.save()

    CONF.cloud = server_uri
    CONF.connect = connect

    if args.ikecfg:
        conn_credentials['ip_address'] = args.ikecfg

    try:
        vdi_client = VdiClient(**conn_credentials)
    except exceptions.MissingParameterException as err:
        logger.error(err)
        return getattr(os, 'EX_SOFTWARE', 1)

    try:
        vdi_connect(vdi_client)
    except exceptions.VDIServerResponseError as err:
        if err.code not in (exceptions.ErrorCodes.PASSWORD_EXPIRED,
                            exceptions.ErrorCodes.PASSWORD_EXPIRE_WARNING,
                            exceptions.ErrorCodes.PASSWORD_ATTEMPTS_WARNING):
            raise
        msg = _(u"The password needs to be changed.")
        logger.info(msg)
        new_password = getpass.getpass(prompt="New password: ")
        confirm_new_password = getpass.getpass(prompt="Confirm new password: ")

        if new_password != confirm_new_password:
            logger.error(_(u'Passwords do not match. '
                           u'Make sure they are the same.'))
            return getattr(os, 'EX_SOFTWARE', 1)

        try:
            vdi_client.change_password(new_password)
        except exceptions.VDIException as err:
            logger.error(err, exc_info=TRACEBACK_ENABLED)
            return getattr(os, 'EX_SOFTWARE', 1)
        else:
            msg = _(u"Password successfully changed.")
            logger.info(msg)
    except Exception as err:
        logger.error(err, exc_info=TRACEBACK_ENABLED)
        return getattr(os, 'EX_SOFTWARE', 1)

    return getattr(os, 'EX_OK', 0)


def create_parser():
    description = u'TIONIX.VDIclient utility'
    parser = argparse.ArgumentParser(description=description,
                                     conflict_handler='resolve')
    parser.add_argument(
        '--version', action='version', version=MODULE_VERSION)
    parser.add_argument(
        '--username',
        required=False,
        type=str,
        help='User name for connection',
    )
    parser.add_argument(
        '--password',
        required=False,
        type=str,
        help='Password for connection',
    )
    parser.add_argument(
        '--project',
        required=False,
        type=str,
        help='Project for vm instance retrieval',
    )
    parser.add_argument(
        '--domain',
        required=False,
        type=str,
        help='Project domain name',
    )
    parser.add_argument(
        '--server-uri',
        required=False,
        type=str,
        help='Uri for vm instance retrieval',
    )
    parser.add_argument(
        '--connect',
        required=False,
        type=str,
        help='RDP-client connection command template. '
             'Example: xfreerdp /u:{user} /p:{password}  /v:{ip}',
    )
    parser.add_argument(
        '--no-ui',
        action='store_true',
        required=False,
        help='Provide this flag if there is no need to launch GUI',
    )
    parser.add_argument(
        '--store-session',
        action='store_true',
        required=False,
        help='Provide this flag if there is a need to save user, password, '
             'project, domain to session file',
    )
    parser.add_argument(
        '--diagnostic',
        action='store_true',
        help='Run self-diagnostic'
    )
    parser.add_argument(
        '--debug',
        dest='debug',
        action='store_true',
        help='Activate the debug mode for self-diagnostic'
    )
    parser.add_argument(
        '--ikecfg',
        required=False,
        type=str,
        help='Client IP for the use of the IPSec/IKE',
    )
    return parser


if __name__ == '__main__':
    sys.exit(main())
