# coding: utf-8

import os
import webbrowser
# noinspection PyPep8Naming
import Tkinter as tk
import ttk
import tkFont

from tionix_vdi_client.widget_redirector import WidgetRedirector
from tionix_vdi_client.helpers import Colors, WindowsCyrillicKeySims
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.settings import RESOURCES_DIR


class CustomCheckBox(tk.Checkbutton):
    def __init__(self, master, row, column, default_value=True, text='',
                 sticky=tk.E):
        self.value = tk.BooleanVar(value=default_value)
        self.config = {
            'text': ' ' + text,
            'variable': self.value,
            'activebackground': Colors.GREY,
        }
        tk.Checkbutton.__init__(self, master, cnf=self.config,
                                variable=self.value,
                                command=self.state_changed)
        self.grid(row=row, column=column, sticky=sticky, padx=2, pady=5)
        self.listeners = []
        self.callbacks = []

    def add_listener(self, elem):
        self.listeners.append(elem)

    def add_callback(self, elem):
        self.callbacks.append(elem)

    def state_changed(self):
        for elem in self.listeners:
            if hasattr(elem, 'checkbox_event'):
                elem.checkbox_event(self.value.get())
        for elem in self.callbacks:
            elem(self.value.get())


# noinspection PyClassHasNoInit,PyUnusedLocal
class LinkLabel(tk.Label):

    def bind_link(self, link, underline=None):
        """
        Binds link to the element. Adds color, text underline and
        cursor change on hover effect.
        :param link: Link to webpage or valid path to local file.
        :param underline: text underline
        :return: None
        """
        def enter(event):
            self.configure(cursor="hand2")

        def leave(event):
            self.configure(cursor="")

        def click(event):
            webbrowser.open_new(link)

        self.bind("<Enter>", enter)
        self.bind("<Leave>", leave)
        self.bind("<Button-1>", click)
        self.configure(fg=Colors.DARK_BLUE, underline=underline)


class CustomEntry(ttk.Entry):
    def __init__(self, *args, **kwargs):
        ttk.Entry.__init__(self, *args, **kwargs)
        self._rc_menu_install()
        self._bind_events()

    def _rc_menu_install(self):
        self.menu = tk.Menu(self, tearoff=0)
        self.menu.add_command(label="Cut")
        self.menu.add_command(label="Copy")
        self.menu.add_command(label="Paste")

        self.menu.entryconfigure(
            "Cut",
            command=lambda: self.focus_force() or self.event_generate(
                "<<Cut>>"))
        self.menu.entryconfigure(
            "Copy",
            command=lambda: self.focus_force() or self.event_generate(
                "<<Copy>>"))
        self.menu.entryconfigure(
            "Paste",
            command=lambda: self.focus_force() or self.event_generate(
                "<<Paste>>"))

    def _bind_events(self):
        self.bind("<Button-3><ButtonRelease-3>", self.show_menu)
        self.bind("<Control-a>", self.event_select_all)

        # Default tkinter events are not bind to cyrillic chars on Windows
        self.bind("<Control-%s>" % WindowsCyrillicKeySims.A,
                  self.event_select_all)
        self.bind("<Control-%s>" % WindowsCyrillicKeySims.C,
                  lambda event: self.event_generate("<<Copy>>"))
        self.bind("<Control-%s>" % WindowsCyrillicKeySims.X,
                  lambda event: self.event_generate("<<Cut>>"))
        self.bind("<Control-%s>" % WindowsCyrillicKeySims.V,
                  lambda event: self.event_generate("<<Paste>>"))

    # noinspection PyUnusedLocal
    def event_select_all(self, event):
        self.focus_force()
        self.selection_range(0, tk.END)
        return 'break'

    def show_menu(self, e):
        self.tk.call("tk_popup", self.menu, e.x_root, e.y_root)

    def get(self):
        value = ttk.Entry.get(self)
        return value.strip()


class CustomPasswordEntry(CustomEntry):

    def __init__(self, *args, **kwargs):
        self.controller = kwargs.pop('controller')
        CustomEntry.__init__(self, *args, **kwargs)

    def checkbox_event(self, checkbox_value):
        if checkbox_value:
            self.controller.update_vdi_password(self.get())
        else:
            if str(self['state']) != tk.NORMAL:
                self.configure(state=tk.NORMAL)
                self.delete(0, tk.END)


class HelpLabel(object):

    def __init__(self, master, row, col, link=''):

        self.help_msg = _(u'Help')

        # default font doesn't have proper underline
        # so we need to configure it
        self.font = tkFont.Font(master, size=10)
        self.font.configure(underline=1)

        self.help_icon = tk.PhotoImage(
            file=os.path.join(RESOURCES_DIR, 'help.gif'))

        self.help_label = LinkLabel(master)
        self.help_label.configure(text=self.help_msg, underline=1,
                                  font=self.font)
        self.help_label.grid(row=row, column=col, padx=20, sticky=tk.E)
        self.help_label.bind_link(link, underline=True)

        # Calc padding between text and icon
        padding = len(self.help_msg) * 13
        self.help_icon_label = LinkLabel(master)
        self.help_icon_label.configure(image=self.help_icon,
                                       width=20, height=20)
        self.help_icon_label.grid(row=1, column=2, sticky=tk.E, padx=padding)
        self.help_icon_label.bind_link(link)


class Button(tk.Button):
    """Button widget that reacts on Enter hit."""

    def __init__(self, master=None, cnf=None, **kwargs):
        cnf = cnf if cnf is not None else {}
        tk.Button.__init__(self, master, cnf, **kwargs)
        self._bind_enter_command(**kwargs)

    def _bind_enter_command(self, **kwargs):
        command = kwargs.get('command')
        if command:
            self.bind('<Return>', command)

    def configure(self, cnf=None, **kwargs):
        tk.Button.configure(self, cnf, **kwargs)
        self._bind_enter_command(**kwargs)


class ToggleButton(Button):
    """Button with two states."""

    def __init__(self, text, text_alter, on_toggle, off_toggle,
                 master=None, cnf=None, **kwargs):
        self.text = text
        self.text_alter = text_alter
        self.on_toggle = on_toggle
        self.off_toggle = off_toggle
        self.toggled = False
        Button.__init__(self, master, cnf, text=text, command=self.toggle,
                        **kwargs)

    def toggle(self):
        if self.toggled:
            self.on_toggle()
            self['text'] = self.text
        else:
            self.off_toggle()
            self['text'] = self.text_alter
        self.toggled = not self.toggled


class ScrollableText(tk.Text):
    """Text field with a scrollbar on the right hand side."""
    MENU_COPY = u"Copy"  # Non-ASCII characters causes menu crash
    MENU_COPY_ALL = u"Copy All"

    def __init__(self, master=None, cnf=None, msg='', **kwargs):
        cnf = cnf if cnf is not None else {}
        tk.Text.__init__(self, master, cnf, **kwargs)

        self._make_readonly()
        self.insert(tk.INSERT, msg)

        scrollbar = tk.Scrollbar(master)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        scrollbar.config(command=self.yview)

        self.config(yscrollcommand=scrollbar.set)
        self.pack(side=tk.LEFT, fill=tk.BOTH)
        self._rc_menu_install()
        self._bind_hotkey_events()

    def _make_readonly(self):
        """Blocks insert\delete events for field."""
        # Workaround for using text.configure(state=tk.DISABLED)
        # since it's blocks copy event.
        # Because nothing should be easy with Tkinter.
        # https://stackoverflow.com/a/11612656/7645774
        self.redirector = WidgetRedirector(self)
        self.insert = self.redirector.register("insert",
                                               lambda *args, **kwargs: "break")
        self.delete = self.redirector.register("delete",
                                               lambda *args, **kwargs: "break")

    def _rc_menu_install(self):
        self.menu = tk.Menu(self, tearoff=0)
        self.menu.add_command(label=self.MENU_COPY)
        self.menu.add_command(label=self.MENU_COPY_ALL)

        self.menu.entryconfigure(
            "Copy",
            command=lambda: (self.focus_force() or
                             self.event_generate("<<Copy>>"))
        )
        self.menu.entryconfigure(
            "Copy All",
            command=lambda: (self.focus_force() or
                             self.select_all() or
                             self.event_generate("<<Copy>>"))
        )

    def _bind_hotkey_events(self):
        self.bind("<Button-3><ButtonRelease-3>", self._show_menu)
        self.bind("<Control-a>", self._event_select_all)
        self.bind("<Control-A>", self._event_select_all)  # if CapsLock pressed

        # Default tkinter events are not bind to cyrillic chars on Windows
        self.bind("<Control-%s>" % WindowsCyrillicKeySims.A,
                  self._event_select_all)
        self.bind("<Control-%s>" % WindowsCyrillicKeySims.C,
                  lambda event: self.event_generate("<<Copy>>"))

    # noinspection PyUnusedLocal
    def _event_select_all(self, event):
        self.select_all()
        return 'break'

    def select_all(self):
        self.tag_add(tk.SEL, "1.0", tk.END)
        self.mark_set(tk.INSERT, "1.0")
        self.see(tk.INSERT)

    def _show_menu(self, e):
        self.tk.call("tk_popup", self.menu, e.x_root, e.y_root)
