# coding: utf-8

import functools
import json
import logging
import string
import sys

from collections import namedtuple

from tionix_vdi_client import TRACEBACK_ENABLED, messagebox
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.settings import CONF

logger = logging.getLogger(__name__)


class ErrorCodes(object):
    if CONF.language == u'ru':
        SUPPORT_MESSAGE = CONF.contact_support_message_ru
    else:
        SUPPORT_MESSAGE = CONF.contact_support_message_en

    UNEXPECTED_ERROR = 1
    LICENSE_ERROR = 2
    TASK_FAILED = 3
    TASK_TIME_OUT = 4
    NO_AVAILABLE_PROJECT = 5
    AUTH_ERROR = 6
    PASSWORD_EXPIRED = 7
    NON_VDI_PROJECT = 8
    INVALID_BODY = 9
    REQUEST_NOT_FOUND = 10
    REQUEST_ACCESS_FORBIDDEN = 11
    PROJECT_QUOTA_EXCEEDED = 12
    SERVER_CREATE_FAILED = 13
    NO_INSTANCE_IN_PROJECT = 14
    DOMAIN_NOT_FOUND = 15
    USER_NOT_FOUND = 16
    INSTANCE_STATE_ERROR = 17
    INSTANCE_NO_IP = 18
    PASSWORD_EXPIRE_WARNING = 19
    PASSWORD_ATTEMPTS_WARNING = 20
    # 21 number isn't used because of confusion prevention
    KERBEROS_SCRIPT_FILE_PROBLEM = 22
    KERBEROS_SCRIPT_RETURN_1 = 23

    auth_problem_codes = (
        AUTH_ERROR, PASSWORD_EXPIRED, PASSWORD_ATTEMPTS_WARNING,
        PASSWORD_EXPIRE_WARNING
    )

    reset_token_errors = (
        PASSWORD_EXPIRED, PASSWORD_EXPIRE_WARNING, PASSWORD_ATTEMPTS_WARNING
    )

    MESSAGES = {
        UNEXPECTED_ERROR: _(
            u"Unexpected error occurred. ") + SUPPORT_MESSAGE,
        LICENSE_ERROR: _(
            u'License of VDI server {cloud_url} got out. ') + SUPPORT_MESSAGE,
        NO_AVAILABLE_PROJECT: _(
            u'No available project found. ') + SUPPORT_MESSAGE,
        AUTH_ERROR: _(u"Invalid credentials. ") + SUPPORT_MESSAGE,
        PASSWORD_EXPIRED: _(u"The password needs to be changed."),
        NON_VDI_PROJECT: _(u'Not a VDI project: {project_name}.'),
        PROJECT_QUOTA_EXCEEDED: _(
            u'Quota exceeded for project {project_name}. '
            u'Please contact system administrator.'),
        SERVER_CREATE_FAILED: _(
            u'Unable to create instance in project {project_name}. '
        ) + SUPPORT_MESSAGE,
        NO_INSTANCE_IN_PROJECT: _(
            u'No instance found in project {project_name}. '
        ) + SUPPORT_MESSAGE,
        DOMAIN_NOT_FOUND: _(
            u'Domain {domain_name} not found.'),
        USER_NOT_FOUND: _(
            u'User {user_name} not found in domain {domain_name}.'),
        INSTANCE_STATE_ERROR: _(
            u'Instance with id {instance_id} is '
            u'in invalid state: {instance_status}. ') + SUPPORT_MESSAGE,
        INSTANCE_NO_IP: _(
            u'Unable to get IP address of instance with id {instance_id}. '
        ) + SUPPORT_MESSAGE,
        PASSWORD_EXPIRE_WARNING: _(u'Password expires in {seconds} seconds '
                                   u'and needs to be changed.'),
        PASSWORD_ATTEMPTS_WARNING: _(u'Password expired, {attempts} grace '
                                     u'logins remain.'),
        KERBEROS_SCRIPT_FILE_PROBLEM: _(
            u"Couldn't change password because of the KERBEROS.script_path "
            u"file doesn't exist or it's unaccessible. ") + SUPPORT_MESSAGE,
        KERBEROS_SCRIPT_RETURN_1: _(
            u"Couldn't change password because of an error while change "
            u"password process execution. ") + SUPPORT_MESSAGE,
    }


class VDIException(Exception):
    pass


class VDIConnectionError(VDIException):
    pass


class ThreadCancelError(VDIException):
    pass


class VDIServerResponseError(VDIException):
    code = ErrorCodes.UNEXPECTED_ERROR

    def __init__(self, message, code=None, data=None):
        super(VDIServerResponseError, self).__init__(message)
        self.data = data
        self._msg_original = message
        if code:
            self.code = code

    @classmethod
    def from_json(cls, json_str, cloud_url=None):
        error_obj = json.loads(json_str)
        error = error_obj.get(u'error', {})
        code = error.get(u'code')
        data = error.get(u'data')
        message = error.get(u'message')
        if data is None:
            data = {}
        data['cloud_url'] = cloud_url
        err_inst = cls(message, data=data)
        err_inst.code = code
        return err_inst

    @property
    def message(self):
        msg = ErrorCodes.MESSAGES.get(self.code) or self._msg_original
        return safeformat(msg, **self.data)


class ApplicationLogicException(VDIException):
    pass


class MissingParameterException(VDIException):
    pass


class ApplicationConnectionError(VDIException):
    pass


class CallbackException(Exception):
    def __init__(self, *args, **kwargs):
        self.callback = kwargs.pop('callback')
        super(CallbackException, self).__init__(*args)


class WeakPassword(CallbackException):
    pass


AlertMessage = namedtuple(
    'AlertMessage',
    ['exception', 'title']
)

ALERT_MESSAGES = (
    AlertMessage(
        exception=VDIException,
        title=_(u'Connection error'),
    ),
    AlertMessage(
        exception=ValueError,
        title=_(u'Invalid input error'),
    ),
)


def process_exception(err, fall=True, parent=None, display_log=False):
    """Exception handler

    Log error and show alert message
    """
    exc = type(err)
    if isinstance(err, VDIServerResponseError):
        if err.code in ErrorCodes.MESSAGES:
            message = ErrorCodes.MESSAGES.get(err.code)
            err_data = err.data or {}
            message = safeformat(message, **err_data)
        else:
            message = ErrorCodes.MESSAGES.get(ErrorCodes.UNEXPECTED_ERROR)
    else:
        message = err.message or str(err)

    for alert_message in ALERT_MESSAGES:
        if issubclass(exc, alert_message.exception):
            title = alert_message.title
            break
    else:
        title = _(u'Error')

    logger.error(message, exc_info=TRACEBACK_ENABLED)

    # noinspection PyBroadException
    try:
        messagebox.showerror(title, message, parent, display_log)
    except Exception as e:
        logger.debug(unicode(e))
        if sys.platform == 'win32':
            from tionix_vdi_client.winapi import message_box
            message_box(message, title)

    if (isinstance(err, CallbackException) and hasattr(err, 'callback') and
            callable(err.callback)):
        try:
            err.callback()
        except Exception as e:
            logger.error(e.message, exc_info=TRACEBACK_ENABLED)

    if fall:
        sys.exit(1)


def catch_exceptions(fall=True):
    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as err:
                parent = getattr(args[0], 'root', None) if args else None
                # noinspection PyTypeChecker
                process_exception(err, fall, parent)

        return wrapped

    return wrapper


def safeformat(template, **kwargs):
    class SafeDict(dict):
        def __missing__(self, key):
            return ''

    return string.Formatter().vformat(template, (), SafeDict(**kwargs))
