# coding: utf-8
import base64
import functools
import logging
import platform
import re
import sys
import urlparse
from StringIO import StringIO
from contextlib import contextmanager
from random import SystemRandom

from Tkinter import TclError

from requests.auth import to_native_string

from tionix_vdi_client.screen_info import get_monitors_info

logger = logging.getLogger(__name__)


API_URL_PATH = 'vdi/api/'
API_DEFAULT_PORT = '8888'
ALLOWED_SCHEMES = ('http://', 'https://')


class Colors(object):
    GREY = '#d9d9d9'
    DARK_BLUE = "#23527c"


class Localization(object):
    EN = u'en'
    RU = u'ru'

    RU_CODE = u'ru_RU'
    EN_US_CODE = u'en_US'

    locales_dict = {
        RU: RU_CODE,
        EN: EN_US_CODE,
    }

    locales_translate = {
        RU: u'Русский',
        EN: u'English',
    }


def get_key_by_value(dict_obj, val_param):
    for key, value in dict_obj.iteritems():
        if value == val_param:
            return key
    return None


def get_win_info():
    return u'{system} {version}'.format(
        system=platform.system(),
        version=platform.release())


def get_linux_info():
    distname, version, _id = platform.linux_distribution()
    return u'{distname} {version}'.format(
        distname=distname,
        version=version)


def get_os_version():
    """Get operating system version"""
    system = platform.system()
    info_getters = {
        u'Linux': get_linux_info,
        u'Windows': get_win_info,
    }
    info_getter = info_getters.get(system, platform.platform)
    return info_getter()


@contextmanager
def suppress_stdout():
    new_stdout = StringIO()
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    logging.disable(logging.ERROR)
    try:
        sys.stdout = new_stdout
        sys.stderr = new_stdout
        yield
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        logging.disable(logging.NOTSET)


def memoized(func):
    cache = func.cache = {}

    @functools.wraps(func)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]
    return memoizer


class WindowsCyrillicKeySims(object):
    """Enum of cyrillic key symbols to their english keys on a keyboard"""
    A = 'ocircumflex'  # ф
    C = 'ntilde'  # с
    X = 'division'  # ч
    Z = 'ydiaeresis'  # я
    V = 'igrave'  # м
    S = 'ucircumflex'  # ы


def decode(text):
    return text.decode('utf-8')


def character_limit(entry, max_len):
    entry_text = entry.get()
    if entry_text:
        entry.set(entry_text[:max_len])


@memoized
def get_monitors_configuration():
    monitors = get_monitors_info()
    logger.debug(
        u'Got following monitors configuration: %s',
        u', '.join([repr(m) for m in monitors]))
    return monitors


def calculate_center_position(tk_window_root, **conf):
    """Calculate window position centered to current user screen."""
    pad_y = conf.get('pad_y')
    pointer_x, pointer_y = tk_window_root.winfo_pointerxy()
    active_monitor, monitors = None, get_monitors_configuration()

    for monitor in monitors:
        if monitor.contains_point(pointer_x, pointer_y):
            active_monitor = monitor
            break
    if active_monitor:
        w = tk_window_root.winfo_reqwidth()
        h = tk_window_root.winfo_reqheight()
        x = active_monitor.offset_x + (active_monitor.width / 2) - (w / 2)
        y = active_monitor.offset_y + (active_monitor.height / 2) - (h / 2)
        if pad_y:
            y += pad_y
        return '+%d+%d' % (x, y)
    return None


def center_window(tk_window_root, **conf):
    tk_window_root.update()
    geom = calculate_center_position(tk_window_root, **conf)
    if geom:
        tk_window_root.geometry(geom)


def auth_header(user, password):
    auth = to_native_string(
        base64.b64encode('{0}:{1}'.format(
            user.encode('utf-8'), password.encode('utf-8'))))
    return {'Authorization': 'Basic {0}'.format(auth)}


def get_vdi_api_url(url):
    """Get VDIserver API url

    Transforms url as follows:
    test.stand.loc -> http://test.stand.loc:8888/vdi/api/
    test.stand.loc:80 -> http://test.stand.loc:80/vdi/api/
    test.stand.loc:80/vdi/api -> http://test.stand.loc:80/vdi/api/
    http://test.stand.loc -> http://test.stand.loc:8888/vdi/api/
    test.stand.loc:8888 -> http://test.stand.loc:8888/vdi/api/
    http://test.stand.loc/vdi/api -> http://test.stand.loc:8888/vdi/api/
    """
    if not any(url.startswith(scheme) for scheme in ALLOWED_SCHEMES):
        url = ALLOWED_SCHEMES[0] + url

    scheme, netloc, _path, _query, _fragment = urlparse.urlsplit(url)

    netloc_parsed = netloc.split(':')
    if len(netloc_parsed) == 1:
        netloc_parsed.append(API_DEFAULT_PORT)
    netloc = ':'.join(netloc_parsed)

    vdi_api_url = urlparse.urlunsplit((scheme, netloc, API_URL_PATH, '', ''))
    return vdi_api_url


class RequestStatus(object):

    PENDING = 'pending'
    PROCESS = 'process'
    ERROR = 'error'
    COMPLETED = 'completed'

    values = (
        PENDING,
        PROCESS,
        ERROR,
        COMPLETED,
    )

    READY_STATUSES = (ERROR, COMPLETED)


def wait_modal_close(root_window, modal_window):
    def handler():
        try:
            modal_window.focus_force()
        except TclError as err:  # window is likely already destroyed
            logger.debug(err)

    if sys.platform == "win32":
        root_window.wait_window(modal_window)
    else:
        wm_delete_window = 'WM_DELETE_WINDOW'
        wm_take_focus = 'WM_TAKE_FOCUS'
        old_close_handler = root_window.protocol(wm_delete_window)
        old_focus_handler = root_window.protocol(wm_take_focus)

        root_window.protocol(wm_delete_window, handler)
        root_window.protocol(wm_take_focus, handler)

        root_window.wait_window(modal_window)

        root_window.protocol(wm_delete_window, old_close_handler)
        root_window.protocol(wm_take_focus, old_focus_handler)


class PasswordGenerator(object):
    PASSWORD_MIN = 8

    LOWER = "abcdefghijklmnopqrstuvwxyz"
    UPPER = LOWER.upper()
    DIGITS = "0123456789"
    SYMBOLS = "!@#$%^&*"
    LOWER_MIN = UPPER_MIN = DIGITS_MIN = SYMBOLS_MIN = 1

    PASSWORD_PATTERN = re.compile(r'^[a-zA-Z0-9!@#$%^&*]+$')
    UPPER_PATTERN = re.compile(r'[A-Z]')
    DIGIT_PATTERN = re.compile(r'[0-9]')
    SYMBOLS_PATTERN = re.compile(r'[%s]' % SYMBOLS)

    @classmethod
    def generate(cls, password_len=None):
        random = SystemRandom()  # use system generator
        password_len = password_len or cls.PASSWORD_MIN
        parts = []

        upper = random.randint(cls.UPPER_MIN,
                               (password_len - cls.DIGITS_MIN -
                                cls.SYMBOLS_MIN - cls.LOWER_MIN))
        password_len -= upper
        parts.append((upper, cls.UPPER))
        digits = random.randint(cls.DIGITS_MIN,
                                password_len - cls.SYMBOLS_MIN - cls.LOWER_MIN)
        password_len -= digits
        parts.append((digits, cls.DIGITS))
        symbols = random.randint(cls.SYMBOLS_MIN, password_len - cls.LOWER_MIN)
        parts.append((symbols, cls.SYMBOLS))
        lower = password_len - symbols
        parts.append((lower, cls.LOWER))

        password = []
        for n, alphabet in parts:
            for _i in range(n):
                password.append(random.choice(alphabet))
        random.shuffle(password)

        return ''.join(password)

    @classmethod
    def check(cls, password):
        valid = False
        if not (len(password) < cls.PASSWORD_MIN or
                not cls.PASSWORD_PATTERN.match(password) or
                cls.UPPER_MIN and not cls.UPPER_PATTERN.search(password) or
                cls.DIGITS_MIN and not cls.DIGIT_PATTERN.search(password) or
                cls.SYMBOLS_MIN and not cls.SYMBOLS_PATTERN.search(password)):
            valid = True
        return valid
