# noinspection PyPep8Naming
from functools import partial

# noinspection PyPep8Naming
import Tkinter as tk
import ttk

from tionix_vdi_client.helpers import center_window, wait_modal_close
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.log import get_log_msg
from tionix_vdi_client.resources import icons
from tionix_vdi_client.elements import ScrollableText, ToggleButton


class Icons(object):
    WARNING = icons.warning
    ERROR = icons.error
    INFO = icons.info
    QUESTION = icons.question


class Replies(object):
    ABORT = "abort"
    RETRY = "retry"
    IGNORE = "ignore"
    OK = "ok"
    CANCEL = "cancel"
    YES = "yes"
    NO = "no"


class ButtonTypes(object):
    OK = "ok"
    OKCANCEL = "okcancel"
    RETRYCANCEL = "retrycancel"
    YESNO = "yesno"
    YESNOCANCEL = "yesnocancel"


BUTTON_PACKS = {
    ButtonTypes.OK: (
        (_('OK'), Replies.OK),
    ),
    ButtonTypes.OKCANCEL: (
        (_('OK'), Replies.OK),
        (_('Cancel'), Replies.CANCEL),
    ),
    ButtonTypes.RETRYCANCEL: (
        (_('Retry'), Replies.RETRY),
        (_('Cancel'), Replies.CANCEL),
    ),
    ButtonTypes.YESNO: (
        (_('Yes'), Replies.YES),
        (_('No'), Replies.NO),
    ),
    ButtonTypes.YESNOCANCEL: (
        (_('Yes'), Replies.YES),
        (_('No'), Replies.NO),
        (_('Cancel'), Replies.CANCEL),
    ),
}


class MessageBox(object):
    """Custom message box window. Can be used with or without root window"""

    result = None

    def __init__(self, title, message, icon, buttons_type, parent=None,
                 display_log=False, **conf):
        # noinspection PyProtectedMember
        aspect = conf.get('aspect')
        parent = parent or tk._default_root
        top = self.top = tk.Toplevel(parent) if parent else tk.Tk()
        top.withdraw()
        top.resizable(width=False, height=False)
        top.update_idletasks()

        if parent:
            top.transient(parent)

        # noinspection PyProtectedMember
        top.tk.call('wm', 'iconphoto', top._w,
                    tk.PhotoImage(data=icons.favicon))
        top.title(title)

        self.message = message
        self.icon = tk.PhotoImage(data=icon)
        self.buttons_type = buttons_type

        self.body(aspect=aspect)
        buttons_frame = self.buttonbox(**conf)
        if display_log:
            self.add_show_log_frame_button(buttons_frame)

        top.bind('<Escape>', lambda event: top.destroy())
        center_window(top, **conf)
        top.deiconify()
        top.grab_set()
        top.focus_force()
        if parent:
            wait_modal_close(parent, top)
        else:
            top.mainloop()

    def body(self, **conf):
        aspect = conf.get('aspect') or 400
        frame = ttk.Frame(self.top)
        ttk.Label(frame, image=self.icon).pack(side=tk.LEFT, padx=5)
        tk.Message(frame, text=self.message, aspect=aspect).pack(side=tk.LEFT)
        frame.pack(padx=5, pady=5)

    def buttonbox(self, **conf):
        btnbox_right = conf.get('btnbox_right')
        set_focus = conf.get('focus', True)
        frame = ttk.Frame(self.top)
        # noinspection PyTypeChecker
        for label, reply in BUTTON_PACKS.get(self.buttons_type):
            command = partial(self._set_result, reply)
            button = ttk.Button(frame, text=label, command=command)
            if set_focus:
                button.focus_set()
                set_focus = False
            button.pack(side=tk.LEFT, padx=5)
            button.bind(
                "<Key-Return>", lambda event: event.widget.invoke())
            button.bind(
                "<Key-KP_Enter>", lambda event: event.widget.invoke())
        if btnbox_right:
            frame.pack(side=tk.RIGHT, pady=5)
        else:
            frame.pack(padx=5, pady=5)
        return frame

    def add_show_log_frame_button(self, buttons_frame):
        def hide_log_frame():
            log_frame.pack_forget()
            win = self.top
            win.resizable(width=False, height=False)
            topbar_height = 100
            if win.winfo_height() > win.winfo_screenheight() - topbar_height:
                # window was stratched to full screen and can't be simply
                # resized until moved back. So we hide it and immidiately
                # show again.
                win.withdraw()
                win.deiconify()
            win.geometry("")

        def show_log_frame():
            self.top.minsize(self.top.winfo_width(), self.top.winfo_height())
            self.top.resizable(width=True, height=True)
            log_frame.config(width=self.top.winfo_width() - 10,
                             height=self.top.winfo_height() + 50)
            log_frame.pack(fill=tk.BOTH, expand=tk.YES)

        btn = ToggleButton(text=_(u"Error Details"),
                           text_alter=_(u"Hide error Details"),
                           on_toggle=hide_log_frame,
                           off_toggle=show_log_frame,
                           master=buttons_frame)
        btn.pack(side=tk.LEFT, padx=5)

        log_frame = ttk.Frame(self.top)
        log_frame.pack_propagate(False)
        ScrollableText(
            log_frame, bg='white', wrap='word', msg=get_log_msg()
        ).pack(fill=tk.BOTH, expand=tk.YES)

    def _set_result(self, result):
        self.result = result
        self.top.destroy()


def _show(title=None, message=None, icon=None, buttons_type=None, parent=None,
          display_log=False, **conf):
    return MessageBox(
        title=title, message=message, icon=icon,
        buttons_type=buttons_type, parent=parent, display_log=display_log,
        **conf
    ).result


def showinfo(title=None, message=None, parent=None):
    """Show an info message"""
    _show(
        title=title, message=message, icon=Icons.INFO,
        buttons_type=ButtonTypes.OK, parent=parent)


def showwarning(title=None, message=None, parent=None, **conf):
    """Show a warning message"""
    _show(
        title=title, message=message, icon=Icons.WARNING,
        buttons_type=ButtonTypes.OK, parent=parent, **conf)


def showerror(title=None, message=None, parent=None, display_log=False,
              **conf):
    """Show an error message"""
    conf['pad_y'] = 150
    _show(
        title=title, message=message, icon=Icons.ERROR,
        buttons_type=ButtonTypes.OK, parent=parent, display_log=display_log,
        **conf)


def askquestion(title=None, message=None, parent=None):
    """Ask a question"""
    return _show(
        title=title, message=message, icon=Icons.QUESTION,
        buttons_type=ButtonTypes.YESNO, parent=parent)


def askokcancel(title=None, message=None, parent=None):
    """Ask if operation should proceed; return true if the answer is ok"""
    result = _show(
        title=title, message=message, icon=Icons.QUESTION,
        buttons_type=ButtonTypes.OKCANCEL, parent=parent)
    return result == Replies.OK


def askyesno(title=None, message=None, parent=None, focus=True):
    """Ask a question; return true if the answer is yes"""
    result = _show(
        title=title, message=message, icon=Icons.QUESTION,
        buttons_type=ButtonTypes.YESNO, parent=parent, focus=focus)
    return result == Replies.YES


def askyesnocancel(title=None, message=None, parent=None):
    """Ask a question; return true if the answer is yes, None if cancelled."""
    result = _show(
        title=title, message=message, icon=Icons.QUESTION,
        buttons_type=ButtonTypes.YESNOCANCEL, parent=parent)
    if result in (Replies.CANCEL, None):
        return None
    return result == Replies.YES


def askretrycancel(title=None, message=None, parent=None):
    """Ask if operation should be retried; return true if the answer is yes"""
    result = _show(
        title=title, message=message, icon=Icons.QUESTION,
        buttons_type=ButtonTypes.RETRYCANCEL, parent=parent)
    if result in (Replies.CANCEL, None):
        return None
    return result == Replies.RETRY
