# coding: utf-8

import ctypes
import ctypes.util
import logging
import operator
import sys

logger = logging.getLogger(__name__)


class MonitorsInfoError(Exception):
    pass


class Monitor(object):

    def __init__(self, width, height, offset_x=0, offset_y=0):
        self.width = width
        self.height = height
        self.offset_x = offset_x
        self.offset_y = offset_y

    def __repr__(self):
        return 'monitor({}x{}+{}+{})'.format(
            self.width, self.height, self.offset_x, self.offset_y)

    def contains_point(self, x, y):
        fits_offset_x = operator.gt if self.offset_x else operator.ge
        fits_offset_y = operator.gt if self.offset_y else operator.ge
        total_width = self.offset_x + self.width
        total_height = self.offset_y + self.height
        return (total_width >= x and fits_offset_x(x, self.offset_x) and
                total_height >= y and fits_offset_y(y, self.offset_y))


def _get_monitors_linux():
    def load_library(name):
        path = ctypes.util.find_library(name)
        if not path:
            raise MonitorsInfoError('Could not load ' + name)
        return ctypes.cdll.LoadLibrary(path)

    class XineramaScreenInfo(ctypes.Structure):
        _fields_ = [
            ('screen_number', ctypes.c_int),
            ('x', ctypes.c_short),
            ('y', ctypes.c_short),
            ('width', ctypes.c_short),
            ('height', ctypes.c_short),
        ]

    xlib = load_library('X11')
    xlib.XOpenDisplay.argtypes = [ctypes.c_char_p]
    xlib.XOpenDisplay.restype = ctypes.POINTER(ctypes.c_void_p)
    d = xlib.XOpenDisplay(b'')
    if not d:
        raise MonitorsInfoError('Could not open display')

    xinerama = load_library('Xinerama')
    if not xinerama.XineramaIsActive(d):
        raise MonitorsInfoError('Xinerama is not active')

    number = ctypes.c_int()
    xinerama.XineramaQueryScreens.restype = (
        ctypes.POINTER(XineramaScreenInfo))
    infos = xinerama.XineramaQueryScreens(d, ctypes.byref(number))
    infos = ctypes.cast(
        infos, ctypes.POINTER(XineramaScreenInfo * number.value)).contents

    return [Monitor(i.width, i.height, i.x, i.y) for i in infos]


def _get_monitors_windows():
    from ctypes import wintypes
    monitors = []

    # noinspection PyUnusedLocal
    def callback(_monitor, _dc, rect, _data):
        rct = rect.contents
        monitors.append(
            Monitor(
                rct.right - rct.left,
                rct.bottom - rct.top,
                rct.left,
                rct.top))
        return True

    # noinspection PyPep8Naming
    MonitorEnumProc = ctypes.WINFUNCTYPE(
        ctypes.c_int,
        ctypes.c_ulong,
        ctypes.c_ulong,
        ctypes.POINTER(wintypes.RECT),
        ctypes.c_double)

    ctypes.windll.user32.EnumDisplayMonitors(
        0, 0, MonitorEnumProc(callback), 0)
    return monitors


def get_monitors_info():
    monitors = []
    if sys.platform == 'win32':
        monitors_getter = _get_monitors_windows
    else:
        monitors_getter = _get_monitors_linux

    try:
        monitors = monitors_getter()
    except MonitorsInfoError as err:
        logger.debug(
            u'Failed to get monitors information: %s', err)
    return monitors
