# coding: utf-8
# pylint: disable=import-error,no-name-in-module

import logging
import os
import sys

from datetime import datetime
from OpenSSL.crypto import load_certificate, FILETYPE_ASN1
from PyKCS11 import (
    PyKCS11Lib, PyKCS11Error,
    CKF_SERIAL_SESSION, CKF_RW_SESSION, CKA_CLASS,
    CKO_CERTIFICATE, CKO_DATA, CKA_TOKEN, CKA_LABEL, CKA_VALUE,
    CKR_PIN_INCORRECT, CKA_PRIVATE, CK_TRUE)
from pyasn1.codec.der.decoder import decode
from pyasn1.type.univ import ObjectIdentifier, Sequence
from tionix_vdi_client import messagebox
from tionix_vdi_client.smart_card.gui import CardViewsController
from tionix_vdi_client.smart_card.pcsc import PCSCError, PCSCClient
from tionix_vdi_client.settings import CONF
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.smart_card.exceptions import (
    CertificateNotFound, SmartCardNotConfigured)

logger = logging.getLogger(__name__)

controller = CardViewsController()


def retrieve_user_login(session, oid=None):
    user_email = find_user_email_from_certificate(session, oid)
    return user_email.split('@')[0].decode('utf-8') if user_email else None


def get_objects(session, template):
    try:
        vdi_files = session.findObjects(template)
    except PyKCS11Error as err:
        logger.warning(u'Error while searching files on smart card: %s.', err)
        vdi_files = []
    return vdi_files


def delete_object(session, obj):
    try:
        session.destroyObject(obj)
    except PyKCS11Error as err:
        logger.warning(u'Error while deleting file on smart card: %s', err)


def get_password_template(login):
    if isinstance(login, unicode):
        login = login.encode('utf-8')
    template = [
        (CKA_CLASS, CKO_DATA),
        (CKA_TOKEN, CK_TRUE),
        (CKA_LABEL, login),
    ]
    return template


def update_password(session, login, new_value):
    template = get_password_template(login)
    vdi_files = get_objects(session, template)
    if vdi_files:
        vdi_file = vdi_files[0]
        delete_object(session, vdi_file)

    store_template = template + [
        (CKA_VALUE, new_value.encode('utf-8')),
        (CKA_PRIVATE, CK_TRUE)
    ]
    return create_object(session, store_template)


def create_object(session, template):
    try:
        result = session.createObject(template)
    except PyKCS11Error as err:
        logger.warning(u'Error while creating file on smart card: %s.', err)
        result = None
    return result


def get_oids_from_x509cert(x509_cert):
    oids = []
    try:
        for i in range(0, x509_cert.get_extension_count()):
            ext = x509_cert.get_extension(i)
            if ext.get_short_name() != 'extendedKeyUsage':
                continue
            eku = ext.get_data()
            for eku_val in decode(eku):
                if not isinstance(eku_val, Sequence):
                    continue
                # noinspection PyProtectedMember
                for v in eku_val._componentValues:
                    if isinstance(v, ObjectIdentifier):
                        oids.append(v.prettyPrint())
    except Exception as err:
        logger.warning(u'Failed to parse certificate OIDs: %s', err)
    return oids


def find_user_email_from_certificate(session, oid=None):
    result_email = None
    certificates = get_objects(session, [(CKA_CLASS, CKO_CERTIFICATE)])

    if not certificates:
        raise CertificateNotFound()

    extracting_attrs = [CKA_VALUE]
    recent_cert_date = None
    recent_cert_oids = []
    now = datetime.utcnow()

    logger.debug(u'Found %d certificates', len(certificates))
    for k, cert in enumerate(certificates, 1):
        logger.debug(u'Processing certificate #%s', k)
        # noinspection PyBroadException
        try:
            attributes = session.getAttributeValue(cert, extracting_attrs)
            attributes = dict(list(zip(extracting_attrs, attributes)))
            x509_cert = load_certificate(
                FILETYPE_ASN1, str(bytearray(attributes[CKA_VALUE])))
        except Exception as err:
            logger.debug(u'Failed to load certificate: %s', err)
            continue

        oids = get_oids_from_x509cert(x509_cert)
        if oid and oid not in oids:
            logger.debug(
                u'Certificate does not contain required EKU OID %s', oid)
            continue

        subject_email = x509_cert.get_subject().emailAddress
        issuer_email = x509_cert.get_issuer().emailAddress

        if subject_email != issuer_email:
            end_date = datetime.strptime(
                x509_cert.get_notAfter(), '%Y%m%d%H%M%SZ')
            if end_date < now:
                logger.debug(u'Certificate expired')
                continue
            start_date = datetime.strptime(
                x509_cert.get_notBefore(), '%Y%m%d%H%M%SZ')
            if not recent_cert_date or start_date > recent_cert_date:
                recent_cert_date = start_date
                result_email = subject_email
                recent_cert_oids = oids

    if result_email is None:
        logger.warning(u'User certificate not found.')
    else:
        logger.debug(
            u'Got the following certificate: '
            u'email=%s, start_date=%s, OIDs=%s',
            result_email, recent_cert_date, recent_cert_oids)
    return result_email


def find_user_password(session, login):
    result_password = None
    template = get_password_template(login)
    vdi_files = get_objects(session, template)
    if vdi_files:
        vdi_file = vdi_files[0]
        try:
            attributes = session.getAttributeValue(vdi_file, [CKA_VALUE])
            result_password = str(bytearray(attributes[0])).decode('utf-8')
        except PyKCS11Error as err:
            logger.warning(
                u'Error while reading vdi file: %s.', err)
    return result_password


# pylint:disable=too-many-statements
def get_session_data():

    if CONF.PYKCS11LIB and "PYKCS11LIB" not in os.environ:
        if os.path.exists(CONF.PYKCS11LIB):
            os.environ["PYKCS11LIB"] = CONF.PYKCS11LIB.encode('utf-8')
        else:
            logger.warning(
                u"PYKCS11LIB file is not found: %s", CONF.PYKCS11LIB)
            raise SmartCardNotConfigured()

    pkcs11 = PyKCS11Lib()

    try:
        pkcs11.load()
    except PyKCS11Error as err:
        logger.warning(
            u'Couldn"t load pkcs11 lib: %s. Please check "PYKCS11LIB" option'
            u' in the configuration file', err)
        raise SmartCardNotConfigured()

    logger.debug(u'PKCS11 lib initialized.')

    slots = pkcs11.getSlotList()
    if not slots:
        logger.warning(u'No card readers found. Login/Pass will be '
                       u'searched in config files.')
        raise SmartCardNotConfigured()

    slot = slots[0]
    logger.info(u'Founded card reader (slot): %s',
                pkcs11.getSlotInfo(slot).slotDescription)

    token_info = pkcs11.getTokenInfo(slot)
    error_msg = None
    min_pin_len = token_info.ulMinPinLen
    max_pin_len = token_info.ulMaxPinLen

    smartcard_error_msg = (
        u"Couldn't connect to smart card. "
        u"Login/Pass will be searched in config files. "
        u"Error: %s")

    try:
        session = pkcs11.openSession(
            slot, CKF_SERIAL_SESSION | CKF_RW_SESSION)
    except PyKCS11Error as err:
        logger.warning(smartcard_error_msg, err)
        raise SmartCardNotConfigured()

    while True:
        try:
            with PCSCClient() as pcsc:
                attempts = pcsc.get_pin_attempts()
        except PCSCError as err:
            logger.warning(u'Failed to get pin attempts: %s', err)
            attempts = '?'

        if not attempts:
            messagebox.showwarning(
                title=_('Smart card authentication'),
                message=_("PIN tries exceeded. Smart card is locked."),
                aspect=3500, btnbox_right=True)
            sys.exit(0)

        controller.show_pin_view(
            error_msg=error_msg, attempts=attempts, min_len=min_pin_len,
            max_len=max_pin_len, win_size="365x90")
        pin = controller.pin

        if pin is None:
            logger.warning(u'Smart card authentication aborted.')
            sys.exit(0)

        # pylint: disable=no-member
        pin = pin.encode('utf-8')

        try:
            session.login(pin)

            try:
                login = retrieve_user_login(session, oid=CONF.SMARTCARD_OID)
            except CertificateNotFound:
                messagebox.showerror(
                    title=_(u'Smart card authentication'),
                    message=_(
                        u'Unable to find certificate on the smart card. '
                        u'Please contact system administrator.'), aspect=3500)
                sys.exit(os.EX_SOFTWARE)

            if not login:
                messagebox.showerror(
                    title=_(u'Smart card authentication'),
                    message=_(
                        u'Unable to get login from the certificate. '
                        u'Please contact system administrator.'), aspect=3500)
                sys.exit(os.EX_SOFTWARE)

            password = find_user_password(session, login)

        except PyKCS11Error as err:
            if err.value == CKR_PIN_INCORRECT:
                error_msg = _(u"You've entered an incorrect PIN.")
                logger.error(error_msg)
                continue
            else:
                logger.warning(smartcard_error_msg, err)
                raise SmartCardNotConfigured()
        else:
            break

    return login, password, session
