# coding: utf-8
from __future__ import unicode_literals

import os

# noinspection PyPep8Naming
import Tkinter as tk

from tionix_vdi_client.helpers import Colors, character_limit
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.settings import RESOURCES_DIR
from tionix_vdi_client.resources.icons import favicon
from tionix_vdi_client.helpers import center_window


class CardViewsController(object):
    pin = None
    password = None
    skip_smartcard = None

    def show_pin_view(self, error_msg=None, attempts=None, min_len=None,
                      max_len=None, win_size=None):
        view = PinWindow(
            error_msg=error_msg, attempts=attempts,
            min_len=min_len, max_len=max_len, win_size=win_size)
        self.pin = view.result


# noinspection PyTypeChecker
class BaseSecureInputWindow(object):
    field_name = None
    window_title = None
    ok_initial_state = tk.NORMAL
    result = None

    def __init__(self, root=None, min_len=None, max_len=None,
                 error_msg='', **kwargs):
        win_size = kwargs.get('win_size')
        self.root = root or tk.Tk()
        self.root.withdraw()
        if win_size:
            self.root.geometry(win_size)
        self.root.title(self.window_title)
        self.root.resizable(width=tk.TRUE, height=tk.FALSE)
        self._set_icon()

        self.min_len = min_len
        self.max_len = max_len

        self.root.grid_columnconfigure(0, minsize=120)
        input_label = tk.Label(self.root, text=self.field_name)
        input_label.grid(row=0, column=0, sticky=tk.W, padx=5, pady=5)

        input_group = tk.Frame(self.root)
        input_group.grid(row=0, column=1, padx=5, pady=5)
        self.entry = self.create_input(self.root)
        self.entry.configure(show='*', takefocus='1')
        self.entry.focus_set()
        self.entry.pack(in_=input_group, side=tk.RIGHT)

        self.raw_password = False
        self.show_pass_btn = tk.Button(input_group)
        self.show_pass_btn.configure(activebackground=Colors.GREY,
                                     command=self.on_show_password_clicked)
        self.eye_opened = tk.PhotoImage(
            file=os.path.join(RESOURCES_DIR, "opened.gif"))
        self.eye_crossed = tk.PhotoImage(
            file=os.path.join(RESOURCES_DIR, "crossed.gif"))
        self.show_pass_btn.pack(
            in_=input_group, side=tk.RIGHT, before=self.entry)

        message_label = tk.Label(self.root, text=error_msg, fg="red")
        message_label.grid(row=1, column=0, columnspan=2)

        buttons_group = tk.Frame(self.root)
        buttons_group.grid(
            row=2, column=0, padx=5, pady=5, sticky=tk.W)

        self.b_ok = tk.Button(
            self.root, text=_('OK'), command=self.on_ok_click,
            state=self.get_ok_state())
        self.b_ok.pack(in_=buttons_group, side=tk.RIGHT, padx=5)

        b_cancel = tk.Button(self.root, text=_('Cancel'),
                             command=self.on_cancel_clicked)
        b_cancel.pack(
            in_=buttons_group, side=tk.RIGHT, padx=0, before=self.b_ok)

        attempts = kwargs.get('attempts')
        if attempts:
            attempts_label = tk.Label(
                self.root,
                text=_('Attempts left: {0}').format(attempts))
            if attempts < 4 or attempts == '?':
                attempts_label.configure(fg="red", font=(None, 8, 'bold'))
            attempts_label.grid(
                row=2, column=1, padx=5, pady=5, sticky=tk.E)

        self.root.bind('<Return>', self.on_ok_click)
        self.root.bind('<KP_Enter>', self.on_ok_click)
        self.root.bind('<Escape>', self.on_cancel_clicked)
        self.root.protocol("WM_DELETE_WINDOW", self.on_cancel_clicked)

        self.toogle_image()
        self.root.resizable(width=tk.FALSE, height=tk.FALSE)
        center_window(self.root)
        self.root.deiconify()
        self.root.mainloop()

    def _set_icon(self):
        img = tk.PhotoImage(data=favicon)
        # noinspection PyProtectedMember
        self.root.tk.call('wm', 'iconphoto', self.root._w, img)

    def toogle_image(self):
        image = self.eye_crossed if self.raw_password else self.eye_opened
        self.show_pass_btn.configure(image=image, width="12", height="12")

    def on_show_password_clicked(self):
        self.raw_password = not self.raw_password
        show = '' if self.raw_password else '*'
        self.entry.configure(show=show)
        self.toogle_image()

    def is_valid(self, text):
        valid = True
        if self.min_len is not None:
            valid = len(text) >= self.min_len

        return valid

    def on_entry_update(self, entry):
        if self.max_len is not None:
            character_limit(entry, self.max_len)

    def create_input(self, root):
        entry_var = tk.StringVar()
        entry_var.trace('w', lambda *args: self.on_entry_update(entry_var))
        return tk.Entry(root, textvariable=entry_var)

    # noinspection PyUnusedLocal
    def on_cancel_clicked(self, event=None):
        self.result = None
        self.root.destroy()

    # noinspection PyUnusedLocal
    def on_ok_click(self, event=None):
        text = self.entry.get()
        if self.is_valid(text):
            self.result = text
            self.root.destroy()

    def get_ok_state(self):
        text = self.entry.get()
        return tk.NORMAL if self.is_valid(text) else tk.DISABLED


class PinWindow(BaseSecureInputWindow):

    field_name = _('Enter PIN')
    window_title = _('Smart card authentication')

    def on_entry_update(self, entry):
        super(PinWindow, self).on_entry_update(entry)
        entry_text = entry.get()
        state = tk.NORMAL if self.is_valid(entry_text) else tk.DISABLED
        self.b_ok.configure(state=state)
