# coding: utf-8
# pylint: disable=too-many-lines
from __future__ import unicode_literals

from collections import defaultdict
import datetime
import logging
import threading

from Queue import Queue
from sys import platform as _platform

# noinspection PyPep8Naming
import Tkinter as tk
import ttk
# noinspection PyPep8Naming

from tionix_vdi_client import MODULE_VERSION, WIN_OS_BITNESS
from tionix_vdi_client import elements, exceptions

from tionix_vdi_client.helpers import (
    Localization, get_key_by_value, Colors, character_limit,
    center_window, calculate_center_position, RequestStatus, wait_modal_close,
    PasswordGenerator
)
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client import messagebox
from tionix_vdi_client.resources.icons import favicon, opened, crossed
from tionix_vdi_client.settings import (
    ConfigSettings, SessionSettings, CONF, REQUEST_REFRESH_TIME
)

from tionix_vdi_client.vdi_connect import VdiClient, rdp_connect

logger = logging.getLogger(__name__)

MAX_TEXT_LENGTH = 60
COPYRIGHTS_TEXT = '© Copyright 2015-%s' % datetime.date.today().year
COMPANY = "TIONIX"
TIONIX_WEB_LINK = 'http://tionix.ru'
HELP_LINK = (
    'http://docs.tionix.ru/tionix_modules/tionix-vdi-client-docs/index.html')

VARIABLES_TITLES = {
    ConfigSettings.CLOUD: _('Primary server URI'),
    ConfigSettings.CONNECT: _('RDP-client connection command.'),
    ConfigSettings.SECONDARY_CLOUD: _('Secondary server URI'),
    ConfigSettings.DOMAIN_NAME: _('Domain name'),
    ConfigSettings.LANGUAGE: _('Language'),
    SessionSettings.PASSWORD: _('Current password'),
}


class TionixVDIClientController(object):
    def __init__(self, card_session=None):
        self.card_session = card_session
        self.view = None
        self.main_root = None
        self.vdi_client = None
        self.ui_mode = False
        self.own_password = False
        self.close = False

    @exceptions.catch_exceptions(fall=True)
    def show_main_window(self, *args, **kwargs):
        """Starting point when module is the main routine."""
        # noinspection PyAttributeOutsideInit
        self.main_root = tk.Tk()
        self.ui_mode = True
        if _platform == "win32":
            self.main_root.tk.call('encoding', 'system', 'cp1252')
        self.view = TionixVDIClientView(self.main_root, self,
                                        *args, **kwargs)
        self.main_root.mainloop()

    def show_change_password_window(self, master_root=None, **kwargs):
        return self._show_view(ResetPasswordWindow, master_root, **kwargs)

    def show_connect_window(self, master_root=None, **kwargs):
        return self._show_view(ConnectWindow, master_root, **kwargs)

    def show_settings_window(self, master_root=None):
        return self._show_view(SettingsWindow, master_root)

    def update_vdi_password(self, new_value):
        self.vdi_client.change_password(new_value)
        if self.card_session:
            try:
                from tionix_vdi_client.smart_card import api  # nopep8
            except ImportError as err:
                logger.warning(
                    u"Environment is not configured: %s. "
                    u"Couldn't save password to smart card",
                    err)
            else:
                login = self.vdi_client.auth_params.get('user')
                result = api.update_password(
                    self.card_session, login, new_value)
                if not result:
                    logger.warning(u'Error while updating vdi password.')

    def _show_view(self, view_cls, master_root, **kwargs):
        master_root = master_root or self.main_root
        root = tk.Toplevel(master_root) if master_root else tk.Tk()
        root.withdraw()
        root.update_idletasks()

        if master_root:
            root.transient(master_root)

        view = view_cls(root, controller=self, **kwargs)
        root.grab_set()
        if master_root:
            wait_modal_close(master_root, root)
        else:
            root.mainloop()
        return getattr(view, 'result', None)


class BaseView(object):
    """Base class for views."""

    # noinspection PyUnusedLocal
    def __init__(self, master, controller, context=None, **kwargs):
        self.root = master
        self.controller = controller
        self.context = context or {}
        self.config_settings = CONF


# noinspection PyPep8Naming,PyUnusedLocal,PyTypeChecker
class TionixVDIClientView(BaseView):
    def __init__(self, *args, **kwargs):
        super(TionixVDIClientView, self).__init__(*args, **kwargs)
        self.smart_card_login = kwargs.get('login')
        self.smart_card_password = kwargs.get('password')

        self.session_settings = SessionSettings()
        self.project_show_entry = self.config_settings.project
        self.show_settings = self.config_settings.show_settings
        self.root.withdraw()
        self.ip_address = kwargs.get('ip_address')

        if self.config_settings.store_session:
            self.session_settings.read()

        self.timer_to_enable = None
        self.raw_passwd = False
        self.root.configure(takefocus='0')
        self.root.title(_('TIONIX.VDIclient'))
        self.root.resizable(width=tk.FALSE, height=tk.FALSE)

        self._configure_grid()
        self._bind_events()
        self._configure_elements()
        self._set_icon()
        center_window(self.root)
        self.root.deiconify()

    @staticmethod
    def _get_element_row(element_type):
        _elements_row = {
            'help': 1,
            'user': 2,
            'password': 3,
            'project': 4,
            'save_parameters': 5,
            'settings': 6,
            'buttons_group': 6,
            'version': 7,
            'copyrights': 8,
        }
        return _elements_row.get(element_type, 0)

    @staticmethod
    def _get_element_column(element_type):
        _elements_columns = dict(label=1, entry=2, help=2)
        return _elements_columns.get(element_type, 0)

    def _configure_grid(self):
        self.root.grid_rowconfigure(self._get_element_row('user'), pad=25)
        self.root.grid_rowconfigure(self._get_element_row('password'), pad=5)
        self.root.grid_rowconfigure(self._get_element_row('project'), pad=25)
        self.root.grid_rowconfigure(
            self._get_element_row('buttons_group'), pad=5)
        self.root.grid_rowconfigure(self._get_element_row('version'), pad=5)

        self.root.grid_columnconfigure(self._get_element_column('label'),
                                       minsize=140)

    def _bind_events(self):
        self.root.bind_class("Button", "<Key-Return>",
                             lambda event: event.widget.invoke())
        self.root.bind_class("Button", "<Key-KP_Enter>",
                             lambda event: event.widget.invoke())
        self.root.bind('<Escape>', self.on_esc_event)
        self.root.protocol("WM_DELETE_WINDOW", self.destroy_window)

    def _configure_elements(self):
        # pylint: disable=too-many-statements
        self.HelpLabel = elements.HelpLabel(
            self.root,
            row=self._get_element_row('help'),
            col=self._get_element_column('help'),
            link=HELP_LINK,
        )

        self.UserLabel = tk.Label(self.root)
        self.UserLabel.configure(text=_('Username'))
        self.UserLabel.grid(row=self._get_element_row('user'),
                            column=self._get_element_column('label'),
                            sticky=tk.W, padx=10)

        self.PasswordLabel = tk.Label(self.root)
        self.PasswordLabel.configure(text=_('Password'))
        self.PasswordLabel.grid(row=self._get_element_row('password'),
                                column=self._get_element_column('label'),
                                sticky=tk.W, padx=10)

        self.UserEntry = elements.CustomEntry(self.root)
        self.UserEntry.configure(
            background='white',
            font='TkFixedFont',
            takefocus='1',
            width=30,
        )
        self.UserEntry.grid(
            row=self._get_element_row('user'),
            column=self._get_element_column('entry'),
            sticky=tk.E,
            padx=10,
        )
        self.UserEntry.focus_set()
        login = (self.session_settings.user
                 if self.is_empty_field(self.smart_card_login)
                 else self.smart_card_login)
        self.UserEntry.insert(0, login)

        if _platform == "win32":
            self.PasswordEntry = elements.CustomEntry(self.root)
        else:
            self.PasswordEntry = elements.CustomPasswordEntry(
                self.root, controller=self.controller
            )
        self.PasswordEntry.configure(
            background='white',
            font='TkFixedFont',
            takefocus='1',
            show='*',
            width=30,
        )
        self.PasswordEntry.grid(
            row=self._get_element_row('password'),
            column=self._get_element_column('entry'),
            sticky=tk.E,
            padx=10
        )

        if self.is_empty_field(self.smart_card_password):
            self.PasswordEntry.insert(0, self.session_settings.password)
        else:
            self.PasswordEntry.insert(0, self.smart_card_password)
            self.PasswordEntry.configure(state='readonly')

        self.PasswordButton = tk.Button(self.root)
        self.PasswordButton.configure(activebackground=Colors.GREY)
        self.PasswordButton.configure(command=self.on_show_password_clicked)
        self.PasswordButton.grid(
            row=self._get_element_row('password'),
            column=self._get_element_column('entry'),
            sticky=tk.E, padx=10, pady=10)
        self.eye_opened = tk.PhotoImage(data=opened)
        self.eye_crossed = tk.PhotoImage(data=crossed)

        if self.project_show_entry:
            self.ProjectLabel = tk.Label(self.root)
            self.ProjectLabel.configure(text=_('Project'))
            self.ProjectLabel.grid(
                row=self._get_element_row('project'),
                column=self._get_element_column('label'),
                sticky=tk.W,
                padx=10
            )

            self.ProjectEntry = elements.CustomEntry(self.root)
            self.ProjectEntry.configure(
                background='white',
                font='TkFixedFont',
                takefocus='1',
                width=30,
            )
            self.ProjectEntry.grid(
                row=self._get_element_row('project'),
                column=self._get_element_column('entry'),
                sticky=tk.E,
                padx=10,
            )
            self.ProjectEntry.insert(0, self.session_settings.project)

        default_value = (not self.is_empty_field(self.smart_card_password) or
                         self.config_settings.store_session)

        self.SessionCheckBox = elements.CustomCheckBox(
            self.root,
            row=self._get_element_row('save_parameters'),
            column=self._get_element_column('entry'),
            default_value=default_value,
            text=_('Remember'),
            sticky=tk.W
        )

        if self.show_settings:
            self.SettingsButton = tk.Button(self.root)
            self.SettingsButton.configure(
                activebackground=Colors.GREY,
                command=self.on_settings_clicked,
                takefocus='1',
                text=_('Settings'))
            self.SettingsButton.grid(row=self._get_element_row('settings'),
                                     column=self._get_element_column('label'),
                                     sticky=tk.W + tk.N, padx=10, pady=10)

        self.ButtonsFrame = tk.Frame(self.root)
        self.ButtonsFrame.grid(
            row=self._get_element_row('buttons_group'),
            column=self._get_element_column('entry'),
            sticky=tk.E + tk.N,
            padx=10, pady=10)

        self.ConnectButton = tk.Button(self.root)
        self.ConnectButton.configure(
            activebackground=Colors.GREY,
            command=self.on_connect_clicked,
            takefocus='1',
            text=_('Connect'))
        self.ConnectButton.focus()
        self.ConnectButton.pack(in_=self.ButtonsFrame, side=tk.LEFT, padx=10)

        self.ExitButton = tk.Button(self.root)
        self.ExitButton.configure(
            activebackground=Colors.GREY,
            command=self.destroy_window,
            takefocus='1',
            text=_('Exit'))
        self.ExitButton.pack(in_=self.ButtonsFrame, side=tk.LEFT)

        self.VersionLabel = tk.Label(self.root)
        self.VersionLabel.configure(text=(_('Version {0}{1}')).format(
            MODULE_VERSION, WIN_OS_BITNESS))
        self.VersionLabel.grid(row=self._get_element_row('version'),
                               column=self._get_element_column('label'),
                               columnspan=2)

        self.CopyrightsFrame = tk.Frame(self.root)
        self.CopyrightsFrame.grid(row=self._get_element_row('copyrights'),
                                  column=self._get_element_column('label'),
                                  columnspan=2)
        self.CopyrightLabel = tk.Label(self.root)
        self.CopyrightLabel.configure(text=COPYRIGHTS_TEXT + ',',
                                      font=(None, 8))
        self.CopyrightLabel.grid(row=1, column=1, in_=self.CopyrightsFrame)

        self.WebsiteLabel = elements.LinkLabel(self.root)
        self.WebsiteLabel.configure(text=COMPANY, font=(None, 8))
        self.WebsiteLabel.grid(row=1, column=2, in_=self.CopyrightsFrame)
        self.WebsiteLabel.bind_link(TIONIX_WEB_LINK)

        self.toogle_image()

    @staticmethod
    def is_empty_field(field_value):
        if field_value is None:
            return True

        field_value = field_value.strip()
        return len(field_value) == 0

    # noinspection PyProtectedMember
    def _set_icon(self):
        img = tk.PhotoImage(data=favicon)
        self.root.tk.call('wm', 'iconphoto', self.root._w, img)

    def on_esc_event(self, event):
        self.destroy_window()

    def save_settings(self):
        conn_credentials = self.get_auth_params()
        self.session_settings.update(conn_credentials)
        self.session_settings.save()

    def session_settings_change(self):
        parameters_changed = False
        conn_credentials = self.get_auth_params()

        for key, value in conn_credentials.items():
            if self.session_settings.get(key) != value:
                parameters_changed = True
                break
        return parameters_changed

    def destroy_window(self):
        store_session = self.SessionCheckBox.value.get()

        self.config_settings.store_session = store_session
        self.config_settings.save()

        if store_session and self.session_settings_change():
            save_settings = messagebox.askyesnocancel(
                _('Exit'),
                _('Connection settings are not saved. '
                  'Would you like to save them before exit?'),
                parent=self.root,
            )
            if save_settings:
                self.save_settings()
            # do not exit on "Cancel"
            close_app = True if save_settings is not None else False
        else:
            close_app = messagebox.askyesno(
                _('Exit'),
                _('Are you sure you want to quit?'),
                parent=self.root
            )
        if close_app:
            self.root.destroy()
            self.root = None

    def enable_connect_button(self):
        if self.root:
            self.ConnectButton.configure(state='normal')

    def disable_connect_button(self):
        self.ConnectButton.configure(state='disabled')
        self.timer_to_enable = threading.Timer(5, self.enable_connect_button)
        self.timer_to_enable.start()

    def get_auth_params(self):
        user = self.UserEntry.get()
        password = self.PasswordEntry.get()
        project = None
        if self.project_show_entry:
            project = self.ProjectEntry.get()

        auth_params = {
            'user': user,
            'password': password,
            'project': project,
            'domain_name': self.config_settings.domain_name,
        }
        return auth_params

    def get_validated_auth_params(self):
        conn_credentials = self.get_auth_params()
        user = conn_credentials.get('user')
        password = conn_credentials.get('password')
        project = conn_credentials.get('project')

        if all((not user, not password)):
            message = _('Please make sure all required fields are filled in.')
            raise ValueError(message)
        message = _(
            'Invalid value provided in field: "%s".\n'
            'Please enter a valid value.'
        )
        if not user:
            message = message % self.UserLabel.cget('text')
            raise ValueError(message)
        if not password:
            message = message % self.PasswordLabel.cget('text')
            raise ValueError(message)

        if not project:
            conn_credentials.pop('project')
        return conn_credentials

    @exceptions.catch_exceptions(fall=False)
    def on_connect_clicked(self):
        self.config_settings.store_session = self.SessionCheckBox.value.get()
        self.config_settings.save()

        conn_credentials = self.get_validated_auth_params()

        if self.config_settings.store_session:
            self.session_settings.update(conn_credentials)
        else:
            self.session_settings.clear()
        self.session_settings.save()

        if self.ip_address:
            conn_credentials['ip_address'] = self.ip_address

        self.controller.vdi_client = VdiClient(**conn_credentials)
        vm_ip = self.controller.show_connect_window(
            context={'username': conn_credentials['user']})

        if self.controller.close:
            self.root.destroy()
            return

        self.PasswordEntry.delete(0, tk.END)
        self.PasswordEntry.insert(
            0, self.controller.vdi_client.auth_params['password'])

        if vm_ip:
            client = self.controller.vdi_client
            rdp_params = dict(
                user=client.auth_params['user'],
                password=client.auth_params['password'],
                domain_name=client.params.get('user_domain_name', ''),
                vm_ip=vm_ip)

            self.disable_connect_button()
            rdp_connect(**rdp_params)

    def toogle_image(self):
        if self.raw_passwd:
            self.PasswordButton.config(image=self.eye_crossed,
                                       width="12", height="12")
        else:
            self.PasswordButton.config(image=self.eye_opened,
                                       width="12", height="12")

    def on_show_password_clicked(self):
        if self.raw_passwd:
            self.PasswordEntry.configure(show='*')
            self.raw_passwd = False
        else:
            self.PasswordEntry.configure(show='')
            self.raw_passwd = True

        self.toogle_image()

    @exceptions.catch_exceptions(fall=False)
    def on_settings_clicked(self):
        self.controller.show_settings_window(self.root)


class WidgetLayout(object):
    default_padx = 10
    default_pady = 10

    def __init__(self, row=0, column=0, sticky=None, padx=None, pady=None):
        self.row = row
        self.column = column
        self.sticky = sticky
        self.padx = padx or self.default_padx
        self.pady = pady or self.default_pady


class SettingsWindow(BaseView):
    def __init__(self, *args, **kwargs):
        super(SettingsWindow, self).__init__(*args, **kwargs)
        self.config_settings = CONF
        self.config_settings.read(reread=True)
        self.entry_widgets = {}
        self._configure_widgets()
        self._configure_window()

    def _configure_window(self):
        self.root.title(_('Settings'))
        self.root.resizable(width=tk.FALSE, height=tk.FALSE)

        img = tk.PhotoImage(data=favicon)
        # noinspection PyProtectedMember
        self.root.tk.call('wm', 'iconphoto', self.root._w, img)

        center_window(self.root)
        self.root.deiconify()

    def _configure_widgets(self):
        self.HelpLabel = elements.HelpLabel(
            self.root,
            row=1,
            col=2,
            link=HELP_LINK,
        )

        # Primary server URI
        self.primary_server_label = tk.Label(
            self.root, text=VARIABLES_TITLES[ConfigSettings.CLOUD])
        self.layout_widget(self.primary_server_label, 'primary_server_label')

        primary_server_entry_var = tk.StringVar()
        primary_server_entry_var.trace(
            'w', lambda *args: character_limit(
                primary_server_entry_var, MAX_TEXT_LENGTH))

        self.primary_server_entry = elements.CustomEntry(
            self.root,
            width=MAX_TEXT_LENGTH,
            textvariable=primary_server_entry_var,
        )
        self.layout_widget(self.primary_server_entry, 'primary_server_entry')
        self.primary_server_entry.insert(0, self.config_settings.cloud)

        self.entry_widgets[ConfigSettings.CLOUD] = self.primary_server_entry

        # Secondary server URI
        self.secondary_server_label = tk.Label(
            self.root, text=VARIABLES_TITLES[ConfigSettings.SECONDARY_CLOUD])
        self.layout_widget(
            self.secondary_server_label, 'secondary_server_label')

        secondary_server_entry_var = tk.StringVar()
        secondary_server_entry_var.trace(
            'w', lambda *args: character_limit(
                secondary_server_entry_var, MAX_TEXT_LENGTH))

        self.secondary_server_entry = elements.CustomEntry(
            self.root,
            width=MAX_TEXT_LENGTH,
            textvariable=secondary_server_entry_var,
        )
        self.layout_widget(
            self.secondary_server_entry, 'secondary_server_entry')
        self.secondary_server_entry.insert(
            0, self.config_settings.secondary_cloud)

        self.entry_widgets[
            ConfigSettings.SECONDARY_CLOUD] = self.secondary_server_entry

        # Domain name
        self.domain_label = tk.Label(
            self.root, text=VARIABLES_TITLES[ConfigSettings.DOMAIN_NAME])
        self.layout_widget(self.domain_label, 'domain_label')

        domain_entry_var = tk.StringVar()
        domain_entry_var.trace(
            'w', lambda *args: character_limit(
                domain_entry_var, MAX_TEXT_LENGTH))

        self.domain_entry = elements.CustomEntry(
            self.root,
            width=MAX_TEXT_LENGTH,
            textvariable=domain_entry_var,
        )
        self.layout_widget(self.domain_entry, 'domain_name')
        domain_name = self.config_settings.domain_name
        self.domain_entry.insert(0, domain_name)
        self.entry_widgets[ConfigSettings.DOMAIN_NAME] = self.domain_entry

        # Language
        self.language_label = tk.Label(
            self.root, text=VARIABLES_TITLES[ConfigSettings.LANGUAGE])
        self.layout_widget(self.language_label, 'language_label')

        self.language_entry = ttk.Combobox(
            self.root, values=Localization.locales_translate.values())

        self.layout_widget(self.language_entry, 'language_entry')

        option = Localization.locales_translate.get(
            self.config_settings.language)

        self.language_entry.set(option)
        self.entry_widgets[ConfigSettings.LANGUAGE] = self.language_entry

        # Buttons
        self.cancel_button = elements.Button(
            self.root, text=_('Cancel'), command=self.on_cancel_clicked)
        self.layout_widget(self.cancel_button, 'cancel_button')

        self.save_button = elements.Button(
            self.root, text=_('Save'), command=self.on_save_clicked)
        self.layout_widget(self.save_button, 'save_button')
        self.save_button.focus()

    @staticmethod
    def layout_widget(widget, name):
        widgets_layout = {
            'primary_server_label': WidgetLayout(row=2, column=1, sticky=tk.W),
            'primary_server_entry': WidgetLayout(row=2, column=2, sticky=tk.W),
            'secondary_server_label': WidgetLayout(
                row=3, column=1, sticky=tk.W),
            'secondary_server_entry': WidgetLayout(
                row=3, column=2, sticky=tk.W),
            'domain_label': WidgetLayout(row=4, column=1, sticky=tk.W),
            'domain_name': WidgetLayout(row=4, column=2, sticky=tk.W),
            'language_label': WidgetLayout(row=5, column=1, sticky=tk.W),
            'language_entry': WidgetLayout(row=5, column=2, sticky=tk.W),
            'save_button': WidgetLayout(row=6, column=2, sticky=tk.E),
            'cancel_button': WidgetLayout(row=6, column=1, sticky=tk.W),
        }
        widget_layout = widgets_layout.get(name) or WidgetLayout()
        widget.grid(
            row=widget_layout.row,
            column=widget_layout.column,
            sticky=widget_layout.sticky,
            padx=widget_layout.padx,
            pady=widget_layout.pady
        )

    # noinspection PyUnusedLocal
    def on_cancel_clicked(self, event=None):
        self.root.destroy()
        return "break"

    # noinspection PyUnusedLocal
    @exceptions.catch_exceptions(fall=False)
    def on_save_clicked(self, event=None):
        # update params for saving
        new_language = None
        old_language = self.config_settings.language
        for setting_param in self.entry_widgets:
            if setting_param in ConfigSettings.ordered_file_variables:
                widget = self.entry_widgets.get(setting_param)
                value = widget.get()

                if setting_param == ConfigSettings.LANGUAGE:
                    value = get_key_by_value(Localization.locales_translate,
                                             value)
                setattr(self.config_settings, setting_param, value)
                new_language = self.config_settings.language

        self.config_settings.save()

        if new_language != old_language:
            msg = _('TIONIX.VDIclient should be restarted to '
                    'change localization.')
            messagebox.showinfo(message=msg, parent=self.root)

        self.root.destroy()
        self.root = None

        return "break"


class ResetPasswordWindow(BaseView):

    CURRENT_PASSWORD = "current_password"
    CURRENT_USER = "current_user"
    NEW_PASSWORD = 'new_password'
    CONFIRM_NEW_PASSWORD = 'confirm_new_password'

    VARIABLES_TITLES = {
        CURRENT_PASSWORD: _('Current password'),
        NEW_PASSWORD: _('New password'),
        CONFIRM_NEW_PASSWORD: _('Confirm new password'),
        CURRENT_USER: _('User'),
    }

    result = None

    # noinspection PyTypeChecker
    def __init__(self, *args, **kwargs):
        super(ResetPasswordWindow, self).__init__(*args, **kwargs)
        self.raw_passwd = defaultdict(bool)
        self.reason = kwargs.get('reason')
        self.conn_params = kwargs.get('conn_params', {})
        self.context = kwargs.get('context', {})
        self.entry_widgets = {}
        self.eye_opened = tk.PhotoImage(data=opened)
        self.eye_crossed = tk.PhotoImage(data=crossed)
        self._configure_widgets()
        self._configure_window()

    def _configure_window(self):
        self.root.title(_('Change password'))
        self.root.resizable(width=tk.FALSE, height=tk.FALSE)

        img = tk.PhotoImage(data=favicon)
        # noinspection PyProtectedMember
        self.root.tk.call('wm', 'iconphoto', self.root._w, img)

        self._set_geom()
        self.root.deiconify()

    def _set_geom(self):
        self.root.update()
        geom = calculate_center_position(self.root)
        if geom:
            self.root.geometry(geom)

    def toogle_image(self, entry, button):
        if self.raw_passwd[entry, button]:
            button.config(image=self.eye_crossed, width="12", height="12")
        else:
            button.config(image=self.eye_opened, width="12", height="12")

    def on_show_password_clicked(self, entry, button):
        def wrapped():
            if self.raw_passwd[entry, button]:
                entry.configure(show='*')
                self.raw_passwd[entry, button] = False
            else:
                entry.configure(show='')
                self.raw_passwd[entry, button] = True
            self.toogle_image(entry, button)
        return wrapped

    def _password_fields(self):
        if self.controller.own_password:
            self.PasswordEntry.configure(state='normal')
            self.ConfirmPasswordEntry.configure(state='normal')
        else:
            self.PasswordEntry.configure(state='readonly')
            self.ConfirmPasswordEntry.configure(state='readonly')

    def _enter_own_password(self, value):
        self.controller.own_password = value
        self._password_fields()

    def _configure_widgets(self):
        self.ReasonLabel = tk.Label(self.root)
        self.ReasonLabel.configure(
            text=self.reason)
        self.ReasonLabel.grid(row=1, column=1, columnspan=2, padx=10, pady=10)

        self.CurrentUserLabel = tk.Label(self.root)
        self.CurrentUserLabel.configure(
            text=self.VARIABLES_TITLES[self.CURRENT_USER])
        self.CurrentUserLabel.grid(
            row=2, column=1, sticky=tk.W, padx=10, pady=5
        )

        self.UserNameLabel = tk.Label(self.root)
        self.UserNameLabel.configure(text=self.context.get('username', ''))
        self.UserNameLabel.grid(row=2, column=2, sticky=tk.W, padx=10, pady=5)

        self.CurrentPasswordLabel = tk.Label(self.root)
        self.CurrentPasswordLabel.configure(
            text=self.VARIABLES_TITLES[self.CURRENT_PASSWORD])
        self.CurrentPasswordLabel.grid(row=3, column=1, sticky=tk.W, padx=10)

        self.CurrentPasswordEntry = elements.CustomEntry(self.root)
        auth_params = getattr(
            self.controller.vdi_client, 'auth_params', self.conn_params)
        self.CurrentPasswordEntry.insert(0, auth_params.get('password', ''))

        self.CurrentPasswordEntry.configure(
            background='white',
            font='TkFixedFont',
            takefocus='1',
            show='*',
        )
        self.CurrentPasswordEntry.grid(row=3, column=2, sticky=tk.EW, padx=10)

        self.PasswordLabel = tk.Label(self.root)
        self.PasswordLabel.configure(
            text=self.VARIABLES_TITLES[self.NEW_PASSWORD])
        self.PasswordLabel.grid(row=4, column=1, sticky=tk.W, padx=10, pady=10)

        self.PasswordEntry = elements.CustomEntry(self.root)

        self.PasswordEntry.configure(
            background='white',
            font='TkFixedFont',
            takefocus='1',
            show='*',
        )
        self.PasswordEntry.grid(
            row=4,
            column=2,
            sticky=tk.EW,
            padx=10,
            pady=10
        )

        self.PasswordButton = tk.Button(self.root)
        self.PasswordButton.configure(activebackground=Colors.GREY)
        self.PasswordButton.configure(
            command=self.on_show_password_clicked(self.PasswordEntry,
                                                  self.PasswordButton))
        self.PasswordButton.grid(
            row=4, column=2, sticky=tk.E, padx=10, pady=10)

        self.ConfirmPasswordLabel = tk.Label(self.root)
        self.ConfirmPasswordLabel.configure(
            text=self.VARIABLES_TITLES[self.CONFIRM_NEW_PASSWORD])
        self.ConfirmPasswordLabel.grid(row=5, column=1, sticky=tk.W, padx=10)

        self.ConfirmPasswordEntry = elements.CustomEntry(self.root)

        self.ConfirmPasswordEntry.configure(
            background='white',
            font='TkFixedFont',
            takefocus='1',
            show='*',
        )
        self.ConfirmPasswordEntry.grid(
            row=5,
            column=2,
            sticky=tk.EW,
            padx=10
        )

        self.ConfirmPasswordButton = tk.Button(self.root)
        self.ConfirmPasswordButton.configure(activebackground=Colors.GREY)
        self.ConfirmPasswordButton.configure(
            command=self.on_show_password_clicked(self.ConfirmPasswordEntry,
                                                  self.ConfirmPasswordButton))
        self.ConfirmPasswordButton.grid(
            row=5, column=2, sticky=tk.E, padx=10)

        if self.config_settings.password_generation:
            self.EnterOwnPassword = elements.CustomCheckBox(
                self.root, 6, 2, default_value=self.controller.own_password,
                text=_('Enter own password'), sticky=tk.W)
            self.EnterOwnPassword.add_callback(self._enter_own_password)

        self.cancel_button = elements.Button(self.root,
                                             text=_('Cancel'),
                                             command=self.on_cancel_clicked)
        self.layout_widget(self.cancel_button, 'cancel_button')

        self.change_button = elements.Button(self.root,
                                             text=_('Save'),
                                             command=self.on_change_clicked)
        self.layout_widget(self.change_button, 'change_button')

        if self.config_settings.password_generation:
            self.generate_password()
            self.change_button.focus_set()
        else:
            self.CurrentPasswordEntry.focus_set()

        self.root.protocol("WM_DELETE_WINDOW", self.on_cancel_clicked)

        self.toogle_image(self.PasswordEntry,
                          self.PasswordButton)
        self.toogle_image(self.ConfirmPasswordEntry,
                          self.ConfirmPasswordButton)

    @staticmethod
    def layout_widget(widget, name):
        widgets_layout = {
            'change_button': WidgetLayout(row=7, column=2, sticky=tk.E),
            'cancel_button': WidgetLayout(row=7, column=1, sticky=tk.W),
        }
        widget_layout = widgets_layout.get(name) or WidgetLayout()
        widget.grid(
            row=widget_layout.row,
            column=widget_layout.column,
            sticky=widget_layout.sticky,
            padx=widget_layout.padx,
            pady=widget_layout.pady
        )

    # noinspection PyUnusedLocal
    def on_cancel_clicked(self, event=None):
        self.root.destroy()
        return "break"

    def generate_password(self):
        random_password = PasswordGenerator.generate()
        self.PasswordEntry.delete(0, tk.END)
        self.PasswordEntry.insert(0, random_password)
        self.ConfirmPasswordEntry.delete(0, tk.END)
        self.ConfirmPasswordEntry.insert(0, random_password)
        self._password_fields()

    # noinspection PyUnusedLocal
    @exceptions.catch_exceptions(fall=False)
    def on_change_clicked(self, event=None):
        password = self.CurrentPasswordEntry.get()
        new_password = self.PasswordEntry.get()
        confirm_new_password = self.ConfirmPasswordEntry.get()

        if new_password != confirm_new_password:
            msg = _(u'Passwords do not match. Make sure they are the same.')
            raise ValueError(msg)

        if self.config_settings.password_generation:
            if not PasswordGenerator.check(new_password):
                msg = _(u"The password must contain latin letters in upper "
                        u"(A-Z) and lower (a-z) case, at least 1 digit (0-9) "
                        u"and at least 1 special character (!@#$%^&*).")
                raise exceptions.WeakPassword(
                    msg, callback=self.generate_password)

        if not self.controller.vdi_client:
            self.conn_params['password'] = password
            self.controller.vdi_client = VdiClient(**self.conn_params)
        else:
            self.controller.vdi_client.set_param(password=password)

        self.controller.update_vdi_password(new_password)
        self.result = new_password
        msg = _(u"Password successfully changed.")
        messagebox.showinfo(message=msg, parent=self.root, title='')

        self.root.destroy()
        self.root = None

        return "break"


class ConnectWindow(BaseView):

    result = None
    REQUEST_MSG = _('Request VM for {user}')
    REQUEST_PENDING_MSG = _('Pending request processing...')
    REQUEST_PROCESS_MSG = _('Processing request...')
    VM_BUILD_MSG = _('Building VM...')
    VM_ACTIVATION_MSG = _('Waiting for VM activation...')
    THREAD_CHECK_TIME = 100  # ms
    PROGESS_BAR_SPEED = 20

    # noinspection PyTypeChecker
    def __init__(self, *args, **kwargs):
        super(ConnectWindow, self).__init__(*args, **kwargs)
        self.root.withdraw()
        self.entry_widgets = {}
        self._configure_widgets()
        self._configure_window()
        self.vdi_client = self.controller.vdi_client
        self.connect_thread = None
        self.connect_queue = None
        self.stop_event = None
        self.root.after(self.THREAD_CHECK_TIME, self.connect)
        self._job = None
        self.root.bind('<Escape>', self.on_cancel_clicked)
        self.root.protocol("WM_DELETE_WINDOW", self.on_cancel_clicked)

    def stop_progress(self):
        self.ProgressBar.stop()
        if self.stop_event:
            self.stop_event.set()
        if self.connect_thread:
            self.connect_thread = None

    @staticmethod
    def _connect_async(queue, vdi_client, stop_event):
        cancel_msg = u'Connection thread was canceled.'
        try:
            request_id = vdi_client.request_vm(stop_event=stop_event)
            while not stop_event.is_set():
                server, status = vdi_client.get_vm(
                    request_id, stop_event=stop_event)
                queue.put((server, status, None))
                if status in RequestStatus.READY_STATUSES:
                    break
                stop_event.wait(REQUEST_REFRESH_TIME)
            else:
                logger.debug(cancel_msg)
        except exceptions.ThreadCancelError:
            logger.debug(cancel_msg)
        except Exception as err:
            queue.put((None, RequestStatus.ERROR, err))

    # pylint: disable=too-many-statements
    def connect(self):
        try:
            if not self.connect_thread:
                self.connect_queue = Queue()
                self.stop_event = threading.Event()
                self.connect_thread = threading.Thread(
                    target=self._connect_async,
                    kwargs={
                        'queue': self.connect_queue,
                        'vdi_client': self.vdi_client,
                        'stop_event': self.stop_event
                    }
                )
                self.connect_thread.start()
                self.ProgressBar.start(self.PROGESS_BAR_SPEED)

            while not self.connect_queue.empty():
                server, status, error = self.connect_queue.get()

                if error:
                    raise error

                if status == RequestStatus.PENDING:
                    self.StatusLabel.configure(
                        text=self.REQUEST_PENDING_MSG)

                elif status == RequestStatus.PROCESS:
                    self.StatusLabel.configure(
                        text=self.REQUEST_PROCESS_MSG)
                    if server:
                        server_status = server.get('status')
                        if server_status == 'BUILD':
                            self.StatusLabel.configure(
                                text=self.VM_BUILD_MSG)
                        elif server_status != 'ACTIVE':
                            self.StatusLabel.configure(
                                text=self.VM_ACTIVATION_MSG)

                elif status == RequestStatus.COMPLETED:
                    if not (server and server.get('ip')):
                        raise exceptions.VDIServerResponseError(
                            _('Failed to get VM ip from server response.')
                        )
                    else:
                        self.result = server.get('ip')
                        break
        except exceptions.VDIServerResponseError as err:
            self.stop_progress()
            change_pwd = True

            parent = self.root
            if self.controller.ui_mode:
                parent = self.root.master

            if err.code in exceptions.ErrorCodes.reset_token_errors:
                reason = err.message
            elif err.code == exceptions.ErrorCodes.AUTH_ERROR:

                if not self.controller.ui_mode:
                    self.controller.vdi_client.auth_params['password'] = ''

                change_pwd = messagebox.askyesno(
                    _('Invalid credentials'),
                    _('If you get temporary password, you could change '
                      'it to a permanent one. Do you want change it now?'),
                    parent=parent, focus=True)
                reason = _('Invalid credentials')
            else:
                self.destroy_window()
                exceptions.process_exception(err, fall=False, display_log=True)
                return

            if change_pwd:
                new_password = self.controller.show_change_password_window(
                    master_root=parent, reason=reason, context=self.context)

                if new_password:
                    self.ProgressBar.start(self.PROGESS_BAR_SPEED)
                    self._job = self.root.after(
                        self.THREAD_CHECK_TIME, self.connect)
                else:
                    self.destroy_window()
            else:
                self.destroy_window()
                self.controller.close = True
        except Exception as err:
            self.destroy_window()
            exceptions.process_exception(err, fall=False, display_log=True)
        else:
            if self.result:
                self.destroy_window()
            else:
                self._job = self.root.after(
                    self.THREAD_CHECK_TIME, self.connect)

    def _configure_window(self):
        self.root.title(_('Request VM'))
        self.root.resizable(width=tk.FALSE, height=tk.FALSE)

        img = tk.PhotoImage(data=favicon)
        # noinspection PyProtectedMember
        self.root.tk.call('wm', 'iconphoto', self.root._w, img)

        center_window(self.root)
        self.root.deiconify()

    def _configure_widgets(self):
        user = self.context.get('username', '')
        self.ProgressBar = ttk.Progressbar(
            self.root, length=300, mode='indeterminate')
        self.ProgressBar.grid(row=1, column=1)

        self.StatusLabel = tk.Label(self.root)
        self.StatusLabel.configure(text=self.REQUEST_MSG.format(user=user),
                                   wraplength=300)
        self.StatusLabel.grid(row=2, column=1)

        self.cancel_button = elements.Button(self.root,
                                             text=_('Cancel'),
                                             command=self.on_cancel_clicked)
        self.cancel_button.grid(row=6, column=1, sticky=tk.E)
        self.cancel_button.focus_set()

    def destroy_window(self):
        if self._job:
            self.root.after_cancel(self._job)
        self.stop_progress()
        self.root.destroy()

    # noinspection PyUnusedLocal
    def on_cancel_clicked(self, event=None):
        result = None
        abort_request = messagebox.askyesno(
            _('Exit'),
            _('Are you sure you want to interrupt VM request?'),
            parent=self.root
        )
        if abort_request:
            self.destroy_window()
            result = "break"
        return result
