# coding: utf-8

import Queue
import binascii
import logging
import os
import subprocess
import threading
import time
from sys import stdout, platform, getfilesystemencoding

import requests

from tionix_vdi_client import CONNECTION_TYPE, CLIENT
from tionix_vdi_client import exceptions
from tionix_vdi_client.helpers import (
    get_os_version, auth_header, get_vdi_api_url, RequestStatus)
from tionix_vdi_client.i18n import ugettext as _, ungettext
from tionix_vdi_client.settings import APP_DIR_PATH, CONF, REQUEST_REFRESH_TIME

logger = logging.getLogger(__name__)

UNEXPECTED_ERR_MSG = _(u"Unexpected error occurred.")


class VDIConnect(object):
    def __init__(self):
        super(VDIConnect, self).__init__()
        self.pipe = []

    def exec_all(self, *args, **kwargs):
        queue = Queue.Queue()
        kwargs['queue'] = queue

        threads_pool = []
        for target in self.pipe:
            thread = threading.Thread(
                target=target, args=args, kwargs=kwargs)
            thread.start()
            threads_pool.append(thread)

        for thread in threads_pool:
            thread.join()

        if not queue.empty():
            err_msg = queue.get()
            raise exceptions.ApplicationLogicException(err_msg)

    def add_exec(self, executer):
        assert callable(executer)
        self.pipe.append(executer)


class BaseExec(object):
    cmd_pipe = []
    bad_cmd_str = _("Invalid parameter: '{connect}'")
    connect_keys = (
        'ip', 'user', 'password', 'project', 'domain', 'rdp_config')

    def __init__(self, **params):
        super(BaseExec, self).__init__()
        self.params = {key: params.get(key, '') for key in self.connect_keys}

    def __call__(self, *args, **kwargs):
        for command in self.cmd_pipe:
            command(*args, **kwargs)


class StdoutExec(BaseExec):
    def __init__(self, **params):
        super(StdoutExec, self).__init__(**params)
        self.cmd_pipe = [self.echo_to_stdout]

    # noinspection PyUnusedLocal
    def echo_to_stdout(self, *args, **kwargs):
        lines = (
            u"username={user}".format(**self.params),
            u"ip={ip}\n".format(**self.params),
        )
        stdout.write('\n'.join(lines))


class VariousStrExec(BaseExec):
    def __init__(self, cmd_str, **params):
        super(VariousStrExec, self).__init__(**params)
        self.cmd_str = cmd_str

    def __call__(self, *args, **kwargs):
        self.queue = kwargs.get('queue')  # pylint:disable=W0201
        keys_to_escape = ['user', 'password', 'domain']
        escaped_params = self.escape_parameters(self.params, keys_to_escape)
        try:
            cmd = self.cmd_str.format(**escaped_params)
        except KeyError:
            msg = "Bad command line format %s" % self.cmd_str
            logger.critical(msg)
            self.queue.put(msg)
            raise exceptions.ApplicationLogicException(self.bad_cmd_str)

        escaped_params.update({'password': '***'})

        logger.debug(u'Running command: %s',
                     self.cmd_str.format(**escaped_params))
        self.run_proc(cmd)

    @staticmethod
    def escape_parameters(parameters, keys_to_escape):
        """
        Escapes specified values in parameters, depending on the OS.
        :param parameters: dict
        :param keys_to_escape: tuple/list
        :return: shallow copy of parameters with modified keys
        """

        def escape_linux(value):
            return "'%s'" % value.replace("'", "'\"'\"'")

        def escape_windows(value):
            return '"%s"' % value

        if platform == "win32":
            escaped = escape_windows
        else:
            escaped = escape_linux

        esc_params = parameters.copy()
        for key in keys_to_escape:
            esc_params[key] = escaped(esc_params[key])

        return esc_params

    def run_proc(self, command):
        process = subprocess.Popen(
            command.encode(getfilesystemencoding()),
            shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        _out, err = process.communicate()
        if process.returncode:
            msg = _(u"Couldn't connect to "
                    u"instance: '{0}'").format(err.decode(stdout.encoding))
            self.queue.put(msg)
        return process


def get_encrypted_rdp_password(raw_password):
    from tionix_vdi_client.winapi import crypt_protect_data

    pass_encrypted = crypt_protect_data(raw_password)
    return binascii.hexlify(pass_encrypted).upper()


def generate_win_rdp_config(host, domain, user, password, ignore_domain):
    user_name = user if ignore_domain else u'{0}\{1}'.format(domain, user)
    rdp_config_template = u"""
    full address:s:{host}
    UserName:s:{user_name}
    password 51:b:{password_encrypted}
    """

    password_encrypted = get_encrypted_rdp_password(password)
    content = rdp_config_template.format(
        host=host,
        domain=domain,
        user_name=user_name,
        password_encrypted=password_encrypted
    )

    rdp_config_path = os.path.join(APP_DIR_PATH, 'connect.rdp')
    with open(rdp_config_path, 'w') as fh:
        fh.write(content.encode('utf-8'))

    return rdp_config_path


def rdp_connect(user, password, domain_name, vm_ip):
    conn_cmd_params = {
        'ip': vm_ip,
        'user': user,
        'password': password,
        'domain': domain_name
    }

    logger.debug(u"Running connection to %s", vm_ip)

    if platform == "win32":
        # on Windows platform some special characters in password can't be
        # escaped correctly, so we generate RDP-file with connection
        # parameters to pass to rdp client
        conn_cmd_params['rdp_config'] = generate_win_rdp_config(
            host=conn_cmd_params['ip'],
            domain=conn_cmd_params['domain'],
            user=conn_cmd_params['user'],
            password=conn_cmd_params['password'],
            ignore_domain=CONF.ignore_domain
        )

    vdi = VDIConnect()
    vdi.add_exec(StdoutExec(**conn_cmd_params))
    vdi.add_exec(VariousStrExec(CONF.connect, **conn_cmd_params))
    vdi.exec_all()


def vdi_connect(vdi_client):
    """For user credentials connect to remote instance"""
    request_id = vdi_client.request_vm()
    while True:
        vm, status = vdi_client.get_vm(request_id)

        if status == RequestStatus.COMPLETED:
            vm_ip = vm['ip']
            break
        else:
            time.sleep(REQUEST_REFRESH_TIME)

    rdp_connect(
        user=vdi_client.auth_params['user'],
        password=vdi_client.auth_params['password'],
        domain_name=vdi_client.params.get('user_domain_name', ''),
        vm_ip=vm_ip)


class VdiClient(object):
    api_url = None

    _reset_token = None

    def __init__(self, user, password, domain_name, project=None,
                 ip_address=None):

        self.auth_params = {
            'user': user,
            'password': password
        }

        self.check_required_values(CONF)
        self.check_required_values(self.auth_params)

        self.params = {}
        if domain_name:
            self.params['user_domain_name'] = domain_name

        if project:
            self.params.update({
                'auth_tenant': project,
                'project_domain_name': domain_name,
            })

        if ip_address:
            self.params['ip_address'] = ip_address
        elif CONF.ikecfg:
            self.params['ip_address'] = CONF.ikecfg

        self.params.update({
            "conn_type": CONNECTION_TYPE,
            "os": get_os_version(),
            "client": CLIENT,
        })

    def set_param(self, **kwargs):
        self.check_required_values(kwargs)
        self.auth_params.update(kwargs)


    @staticmethod
    def check_required_values(parameters):
        """
        Checks required keys for empty values.
        :param parameters: dict or Settings()
        :raises: ApplicationLogicException
        """
        from tionix_vdi_client.ui_tk import VARIABLES_TITLES

        required_keys = getattr(
            parameters, 'required_attributes', parameters.keys())
        for key, value in parameters.items():
            if key in required_keys and value in ('', None):
                raise exceptions.MissingParameterException(
                    (_(u"Parameter \"{0}\" is required")).format(
                        VARIABLES_TITLES.get(key, key))
                )

    # pylint: disable=too-many-statements
    def _action(self, api_method, body=None, extra_params=None,
                stop_event=None):
        headers = auth_header(**self.auth_params)

        if not self.api_url:
            api_urls = [get_vdi_api_url(CONF.cloud)]
            if CONF.secondary_cloud:
                api_urls.append(get_vdi_api_url(CONF.secondary_cloud))
            if CONF.additional_clouds:
                api_urls += [get_vdi_api_url(url)
                             for url in CONF.additional_clouds]
        else:
            api_urls = [self.api_url]

        retries = CONF.retries
        if retries >= 0:
            retries += 1

        params = dict(**self.params)
        if extra_params:
            params.update(extra_params)

        default_err_msg = exceptions.ErrorCodes.MESSAGES[
            exceptions.ErrorCodes.UNEXPECTED_ERROR]
        server_errs = {url: default_err_msg for url in api_urls}

        while retries > 0 or retries == -1:
            if stop_event and stop_event.is_set():
                raise exceptions.ThreadCancelError

            for api_url in api_urls:
                logger.info(u'Performing %s action on %s', api_method, api_url)
                url = u'{0}{1}/'.format(api_url, api_method)

                try:
                    logger.debug(u'Requesting url {0} with parameters: {1}'
                                 u''.format(url, params))
                    if body:
                        response = requests.post(
                            url, params=params, headers=headers, json=body)
                    else:
                        response = requests.get(
                            url, params=params, headers=headers)
                except requests.exceptions.ConnectionError:
                    msg = _(u"Couldn't connect to server: {}").format(url)
                    logger.error(msg)
                    server_errs[api_url] = msg
                except requests.exceptions.InvalidURL:
                    msg = _(u"Invalid address: {}").format(url)
                    logger.error(msg)
                    server_errs[api_url] = msg
                else:
                    try:
                        result = self._parse_response(response, cloud=api_url)
                    except exceptions.VDIServerResponseError as err:
                        exc_codes = exceptions.ErrorCodes.auth_problem_codes
                        if err.code in exc_codes:
                            raise err

                        server_errs[api_url] = err.message
                    else:
                        if result is not None:
                            self.api_url = api_url
                            return result

            if retries > 0:
                retries -= 1
                if retries > 0:
                    logger.info(u"Unable to connect to cloud. Tries left: {0}"
                                u"".format(retries))
            elif retries < 0:
                logger.info(u"Unable to connect to cloud. Retry after {0} "
                            u"seconds.".format(CONF.timeout))
            if retries != 0:
                logger.info(u"Timeout before retry connection to cloud.")
                if stop_event:
                    stop_event.wait(CONF.timeout)
                else:
                    time.sleep(CONF.timeout)

        msg = ungettext(
            u"Unable to connect to server: {}",
            u"Unable to connect to servers: {}",
            len(api_urls)
        )

        fail_reason_msgs = [
            _(u"{server_url}\nReason: {err}").format(server_url=url, err=err)
            for url, err in server_errs.items()
        ]
        raise exceptions.ApplicationConnectionError(
            msg.format(u"\n\n".join(fail_reason_msgs))
        )

    def request_vm(self, stop_event=None):
        """
            Connects to server and retrieves instance IP address
            :returns: ip address
        """

        api_method = 'request-vm'

        response = self._action(api_method, stop_event=stop_event)

        request_id = response.get('request_id')
        if not request_id:
            msg = _(u"Couldn't get request id from server response.")
            logger.error(u"Couldn't get request_id from json: %s", response)
            raise exceptions.VDIServerResponseError(msg)
        logger.debug(u"Got request_id: %s", request_id)
        return request_id

    def get_vm(self, request_id, stop_event=None):
        """
            Connects to server and retrieves instance IP address
            :returns: ip address
        """

        api_method = 'get-vm'
        response = self._action(
            api_method, extra_params={'request_id': request_id},
            stop_event=stop_event)
        server = response.get('server')
        status = response.get('status')
        return server, status

    def change_password(self, new_password, stop_event=None):
        data = {'password': new_password}

        if self._reset_token:
            data['token'] = self._reset_token

        api_method = 'change-password'
        self._action(api_method, body=data, stop_event=stop_event)
        self.auth_params['password'] = new_password

    @staticmethod
    def _get_response_json(response):
        """Return json dict from response"""
        try:
            json_body = response.json()
        except ValueError:
            msg = _(u"Response parsing error.")
            logger.error(
                u"Response doesn't have json. Request url: %s", response.url)
            raise exceptions.VDIServerResponseError(msg)

        logger.debug(u'Got following response: {0}'.format(json_body))
        return json_body

    def _parse_response(self, response, cloud=None):
        result = None
        if response.status_code in (500, 503):
            logger.error(u"Error %d. Url: %s.",
                         response.status_code, response.url)
        elif response.status_code >= 400:
            response_content = response.content.decode('utf-8')
            err_msg = u"Error %d. Content: %s"
            logger.error(err_msg, response.status_code, response_content)

            if response.status_code == 400:
                error = exceptions.VDIServerResponseError.from_json(
                    response_content, cloud_url=cloud)

                if error.code in exceptions.ErrorCodes.reset_token_errors:
                    self._reset_token = error.data.get('token')
                else:
                    self._reset_token = None
                raise error
        else:
            result = self._get_response_json(response)
        return result
